/*
 *  Copyright (c) 2008 Rodrigo Ruiz
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.apache.axis.message.addressing;

import javax.xml.namespace.QName;
import javax.xml.soap.SOAPException;

import org.apache.axis.AxisFault;
import org.apache.axis.MessageContext;
import org.apache.axis.message.MessageElement;
import org.apache.axis.message.SOAPHeaderElement;
import org.apache.axis.soap.SOAPConstants;

/**
 * AddressingHeaderFault type.
 *
 * @author Davanum Srinivas
 * @author Rodrigo Ruiz
 * @version $Revision: 14 $
 */
public class AddressingHeaderFault extends AxisFault {

  /**
   * <code>serialVersionUID</code> attribute.
   */
  private static final long serialVersionUID = -7037754917195807959L;

  /**
   * Namespace used by all constants in this class.
   */
  private static final String NS = Constants.NS_URI_ADDRESSING_2005_08;

  /**
   * Invalid address header fault name.
   */
  private static final QName INVALID_ADDRESSING_HEADER = new QName(NS,
    "InvalidAddressingHeader");

  /**
   * Invalid address fault name.
   */
  public static final QName INVALID_ADDRESS = new QName(NS, "InvalidAddress");

  /**
   * Invalid cardinality fault name.
   */
  public static final QName INVALID_CARDINALITY = new QName(NS, "InvalidCardinality");

  /**
   * Header required fault name.
   */
  public static final QName HEADER_REQUIRED = new QName(NS,
    "MessageAddressingHeaderRequired");

  /**
   * Invalid EPR fault name.
   */
  public static final QName INVALID_EPR = new QName(NS, "InvalidEPR");

  /**
   * Missing address fault name.
   */
  public static final QName MISSING_ADDRESS = new QName(NS, "MissingAddressInEPR");

  /**
   * Duplicate Message ID fault name.
   */
  public static final QName DUPLICATE_MSGID = new QName(NS, "DuplicateMessageID");

  /**
   * Action mismatch fault name.
   */
  public static final QName ACTION_MISMATCH = new QName(NS, "ActionMismatch");

  /**
   * Problem header name.
   */
  public static final QName PROBLEM_HEADER = new QName(NS, "ProblemHeaderQName");

  /**
   * Fault detail field name.
   */
  public static final QName FAULT_DETAIL = new QName(NS, "FaultDetail");

  /**
   * Creates a new instance.
   *
   * @param faultString Error message
   */
  public AddressingHeaderFault(String faultString) {
    this(faultString, null, null);
  }

  /**
   * Creates an instance, using a predefined message depending on the sub-code.
   *
   * @param subSubCode    Error sub-code
   * @param problemHeader Problematic header name
   */
  public AddressingHeaderFault(QName subSubCode, String problemHeader) {
    this(buildMessage(subSubCode, problemHeader), subSubCode, problemHeader);
  }

  /**
   * Creates a new instance.
   *
   * @param faultString Error message
   * @param subSubCode Error code
   * @param problemHeader Problem header
   */
  public AddressingHeaderFault(String faultString, QName subSubCode,
    String problemHeader) {
    MessageContext mc = MessageContext.getCurrentContext();

    if (mc == null || mc.getSOAPConstants() == SOAPConstants.SOAP12_CONSTANTS) {
      this.setFaultCode(org.apache.axis.Constants.FAULT_SOAP12_SENDER);
      if (HEADER_REQUIRED.equals(subSubCode)) {
        this.addFaultSubCode(HEADER_REQUIRED);
      } else {
        this.addFaultSubCode(INVALID_ADDRESSING_HEADER);
        if (subSubCode != null) {
          this.addFaultSubCode(subSubCode);
        }
      }

      if (problemHeader != null) {
        this.addFaultDetail(PROBLEM_HEADER, "wsa:" + problemHeader);
      }
    } else {
      if (subSubCode != null && subSubCode.equals(HEADER_REQUIRED)) {
        this.setFaultCode(HEADER_REQUIRED);
      } else {
        this.setFaultCode(INVALID_ADDRESSING_HEADER);
      }

      if (problemHeader != null) {
        try {
          SOAPHeaderElement detailHeader = new SOAPHeaderElement(FAULT_DETAIL);
          MessageElement problemEl = new MessageElement(PROBLEM_HEADER, "wsa:"
            + problemHeader);
          detailHeader.addChild(problemEl);
          this.addHeader(detailHeader);
        } catch (SOAPException e) {
          // Will never happen, but... Log?
          log.error(e);
        }
      }
    }

    this.setFaultString(faultString);
  }

  /**
   * Creates a new instance.
   *
   * @param faultString Error message
   * @param problemHeader Header error source
   */
  public AddressingHeaderFault(String faultString, String problemHeader) {
    this(faultString, null, problemHeader);
  }

  /**
   * Gets a message that will depend on the specified code and header name.
   *
   * @param code   Error code
   * @param header Header name
   * @return Error message
   */
  private static String buildMessage(QName code, String header) {
    String localName = (code == null) ? INVALID_ADDRESSING_HEADER.getLocalPart()
                                      : code.getLocalPart();
    if ("InvalidCardinality".equals(localName)) {
      return "Duplicate " + header + " header";
    } else {
      return "Error in " + header + " header";
    }
  }
}
