/*
 *  Copyright (c) 2008 Rodrigo Ruiz
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.apache.axis.message.addressing;

import java.util.ArrayList;
import java.util.Iterator;

import javax.xml.namespace.QName;
import javax.xml.soap.SOAPElement;

import org.apache.axis.encoding.AnyContentType;
import org.apache.axis.message.MessageElement;
import org.apache.axis.message.addressing.util.AddressingUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * Java class for complex types like ReferencePropertiesType, MetaDataType, etc.
 * <p>
 * The following schema fragment specifies the expected content contained within
 * this java content object. (defined at
 * http://schemas.xmlsoap.org/ws/2004/08/addressing line 62)
 * <p>
 *
 * <pre>
 * &lt;complexType name=&quot;ReferencePropertiesType&quot;&gt;
 *   &lt;complexContent&gt;
 *     &lt;restriction base=&quot;{http://www.w3.org/2001/XMLSchema}anyType&quot;&gt;
 *       &lt;sequence&gt;
 *         &lt;any/&gt;
 *       &lt;/sequence&gt;
 *     &lt;/restriction&gt;
 *   &lt;/complexContent&gt;
 * &lt;/complexType&gt;
 * </pre>
 *
 * @author Rodrigo Ruiz
 * @version $Revision: 14 $
 * @version $Revision: 14 $
 */
public abstract class AnyContentTypeList extends ArrayList<MessageElement>
  implements AnyContentType, DOMAppendable {

  /**
   * Class logger.
   */
  private static final Log LOG = LogFactory.getLog(AnyContentTypeList.class);

  /**
   * Parses an XML element node and initializes an instance of this class
   * with the extracted data.
   *
   * @param list    Instance to be initialized
   * @param element Element to parse
   */
  protected static void fromElement(AnyContentTypeList list, Element element) {
    NodeList children = element.getChildNodes();
    for (int i = 0; i < children.getLength(); i++) {
      Node child = children.item(i);
      if (!(child instanceof Element)) {
        continue;
      }
      list.add(child);
    }
  }

  /**
   * Parses a SOAP element node and initializes an instance of this class
   * with the extracted data.
   *
   * @param list    Instance to be initialized
   * @param element Element to parse
   */
  protected static void fromSOAPElement(AnyContentTypeList list, SOAPElement element) {
    Iterator<?> iter = element.getChildElements();
    while (iter.hasNext()) {
      javax.xml.soap.Node child = (javax.xml.soap.Node) iter.next();
      if (!(child instanceof SOAPElement)) {
        continue;
      }
      if (child instanceof MessageElement) {
        try {
          list.add(((MessageElement) child).getAsDOM());
        } catch (Exception e) {
          list.add(child);
        }
      } else {
        list.add(child);
      }
    }
  }

  /**
   * Creates a new instance.
   */
  public AnyContentTypeList() {
  }

  /**
   * Creates a new instance, performing a shallow or deep copy of another
   * instance.
   *
   * @param src Elements to copy
   * @param deepCopy Whether to perform a shallow or deep copy
   */
  public AnyContentTypeList(AnyContentTypeList src, boolean deepCopy) {
    if (src == null) {
      throw new IllegalArgumentException("Copy from null instance not allowed");
    } else if (!this.getClass().isAssignableFrom(src.getClass())) {
      throw new IllegalArgumentException("Copy from a " + src.getClass().getName()
        + " instance into a " + this.getClass().getName() + " not allowed");
    }

    if (deepCopy) {
      for (MessageElement elem : src) {
        try {
          this.add(new MessageElement(elem.getAsDOM()));
        } catch (Exception e) {
          // if failed just copy reference
          this.add(elem);
        }
      }
    } else {
      // shallow copy
      this.addAll(src);
    }
  }

  /**
   * Creates a new instance and adds the specified object as an XML element.
   *
   * @param element Item to add
   */
  public AnyContentTypeList(Object element) {
    this.add(element);
  }

  /**
   * Adds an item to the properties.
   *
   * @param item Item to add
   */
  public void add(Object item) {
    MessageElement value = null;
    if (item instanceof MessageElement) {
      value = (MessageElement) item;
    } else if (item instanceof Element) {
      value = new MessageElement((Element) item);
    } else {
      throw new IllegalArgumentException();
    }
    this.add(value);
  }

  /**
   * {@inheritDoc}
   */
  public void append(Element parent, String elementName) {
    append(AddressingUtils.getAddressingVersion(), parent, elementName);
  }

  /**
   * {@inheritDoc}
   */
  public void append(AddressingVersion version, Element parent, String elementName) {
    final int size = this.size();

    if (size == 0) {
      return;
    } else {
      Document doc = parent.getOwnerDocument();

      String ns = version.getNamespace();
      Element elem = doc.createElementNS(ns, elementName);

      try {
        for (MessageElement item : this) {
          elem.appendChild(doc.importNode(item.getAsDOM(), true));
        }
      } catch (Exception e) {
        LOG.warn("Error appending child element", e);
      }
      parent.appendChild(elem);
    }
  }

  /**
   * Get a (single) MessageElement matching the passed QName.
   *
   * @param qname Name to look for
   * @return Message element matching the passed name
   */
  public MessageElement get(QName qname) {
    for (MessageElement elem : this) {
      QName name = elem == null ? null : elem.getQName();
      if (name != null && name.equals(qname)) {
        return elem;
      }
    }
    return null;
  }

  /**
   * {@inheritDoc}
   */
  // CHECKSTYLE:OFF (get_any cannot be renamed, as it is part of the Axis API)
  public MessageElement[] get_any() {
  // CHECKSTYLE:ON
    return this.toArray(new MessageElement[this.size()]);
  }

  /**
   * {@inheritDoc}
   */
  // CHECKSTYLE:OFF (set_any cannot be renamed, as it is part of the Axis API)
  public void set_any(MessageElement[] any) {
  // CHECKSTYLE:ON
    this.clear();
    for (MessageElement elem : any) {
      this.add(elem);
    }
  }

  /**
   * Gets a string representation of this instance.
   *
   * @param itemName Name to use for elements in this instance
   * @return A string containing a description of this object
   */
  protected String toString(String itemName) {
    StringBuffer buf = new StringBuffer();
    buf.append(itemName);

    for (int i = 0; i < this.size(); i++) {
      buf.append(i == 0 ? ": { " : ", ");
      MessageElement item = this.get(i);
      try {
        buf.append(item.getAsString());
      } catch (Exception e) {
        buf.append("<error converting: " + e.getMessage() + ">");
      }
    }
    buf.append(" }");
    return buf.toString();
  }

}
