/*
 *  Copyright (c) 2008 Rodrigo Ruiz
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.apache.axis.message.addressing;

import javax.xml.soap.Name;
import javax.xml.soap.SOAPEnvelope;
import javax.xml.soap.SOAPException;
import javax.xml.soap.SOAPHeader;
import javax.xml.soap.SOAPHeaderElement;

import org.apache.axis.message.addressing.util.AddressingUtils;
import org.apache.axis.message.addressing.util.TextExtractor;
import org.apache.axis.types.URI;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * Java content class for AttributedURI complex type.
 * <p>
 * The following schema fragment specifies the expected content contained within
 * this java content object. (defined at
 * http://schemas.xmlsoap.org/ws/2004/08/addressing line 105)
 *
 * <pre>
 * &lt;complexType name=&quot;AttributedURI&quot;&gt;
 *   &lt;simpleContent&gt;
 *     &lt;extension base=&quot;&lt;http://www.w3.org/2001/XMLSchema&gt;anyURI&quot;&gt;
 *     &lt;/extension&gt;
 *   &lt;/simpleContent&gt;
 * &lt;/complexType&gt;
 * </pre>
 *
 * @author Davanum Srinivas
 * @author Rodrigo Ruiz
 * @version $Revision: 14 $
 */
public class AttributedURI extends URI
  implements DOMAppendable, AddressingHeaderItem {

  /**
   * <code>serialVersionUID</code> attribute.
   */
  private static final long serialVersionUID = -3363815864897359681L;

  /**
   * Constructor AttributedURI.
   */
  public AttributedURI() {
  }

  /**
   * Constructor AttributedURI.
   *
   * @param el Element to extract the URI from
   * @throws URI.MalformedURIException If the extracted URI is not valid
   */
  public AttributedURI(SOAPHeaderElement el) throws URI.MalformedURIException {
    this(TextExtractor.getText(el).trim());
  }

  /**
   * Constructor Address.
   *
   * @param uri URI
   * @throws MalformedURIException If the URI is not valid
   */
  public AttributedURI(String uri) throws MalformedURIException {
    super(new URI(uri.trim()));
  }

  /**
   * Constructor AttributedURI.
   *
   * @param uri URI
   */
  public AttributedURI(URI uri) {
    super(uri);
  }

  /**
   * Gets the default name of the element to create in DOMAppendable and
   * AddressingHeaderItem methods.
   *
   * @return Element name
   */
  public String getDefaultElementName() {
    return "";
  }

  /**
   * {@inheritDoc}
   */
  public final void append(Element parent) {
    append(parent, getDefaultElementName());
  }

  /**
   * {@inheritDoc}
   */
  public final void append(Element parent, String elementName) {
    append(AddressingUtils.getAddressingVersion(), parent, elementName);
  }

  /**
   * {@inheritDoc}
   */
  public final void append(AddressingVersion version, Element parent, String elemName) {
    Document doc = parent.getOwnerDocument();
    Element child = doc.createElementNS(version.getNamespace(), elemName);
    child.appendChild(doc.createTextNode(this.toString()));
    parent.appendChild(child);
  }

  /**
   * Converts this instance into a SOAP header element.
   *
   * @param version  WS-Addressing version to use
   * @param env      Envelope to put the header into
   * @param actorURI Actor URI
   * @return Created element
   * @throws SOAPException If an error occurs
   */
  public final SOAPHeaderElement toSOAPHeaderElement(AddressingVersion version,
    SOAPEnvelope env, String actorURI) throws SOAPException {
    return toSOAPHeaderElement(version, env, actorURI, getDefaultElementName());
  }

  /**
   * Converts this instance into a SOAP header element.
   *
   * @param version  WS-Addressing version to use
   * @param env      Envelope to put the header into
   * @param actorURI Actor URI
   * @param name     Element name
   * @return Created element
   * @throws SOAPException If an error occurs
   */
  protected SOAPHeaderElement toSOAPHeaderElement(AddressingVersion version,
    SOAPEnvelope env, String actorURI, String name) throws SOAPException {

    if (name == null) {
      throw new IllegalArgumentException("A null name is not allowed.");
    }
    Name qname = env.createName(name, Constants.NS_PREFIX_ADDRESSING, version
      .getNamespace());
    SOAPHeader header = env.getHeader();
    if (header == null) {
      header = env.addHeader();
    }

    // add the new header element
    SOAPHeaderElement headerElement = header.addHeaderElement(qname);
    headerElement.setActor(actorURI);
    headerElement.addTextNode(this.toString());
    return headerElement;
  }

  /**
   * {@inheritDoc}
   */
  public final SOAPHeaderElement toSOAPHeaderElement(SOAPEnvelope env, String actorURI)
    throws SOAPException {
    AddressingVersion version = AddressingUtils.getAddressingVersion();
    return this.toSOAPHeaderElement(version, env, actorURI);
  }

}
