/*
 *  Copyright (c) 2008 Rodrigo Ruiz
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.apache.axis.message.addressing;

import java.util.Iterator;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.soap.SOAPElement;
import javax.xml.soap.SOAPEnvelope;
import javax.xml.soap.SOAPException;
import javax.xml.soap.SOAPHeader;
import javax.xml.soap.SOAPHeaderElement;

import org.apache.axis.message.MessageElement;
import org.apache.axis.message.addressing.util.AddressingUtils;
import org.apache.axis.types.URI;
import org.apache.axis.types.URI.MalformedURIException;
import org.apache.axis.utils.XMLUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * Java content class for EndpointReference element declaration.
 * <p>
 * The following schema fragment specifies the expected content contained within
 * this java content object. (defined at
 * http://schemas.xmlsoap.org/ws/2004/08/addressing line 45)
 *
 * <pre>
 * &lt;element
 *   xmlns:wsa=&quot;http://schemas.xmlsoap.org/ws/2004/08/addressing&quot;
 *   name=&quot;EndpointReference&quot;
 *   type=&quot;wsa:EndpointReferenceType&quot;/&gt;
 * </pre>
 *
 * @author Davanum Srinivas
 * @author Rodrigo Ruiz
 * @version $Revision: 13 $
 */
public class EndpointReference extends EndpointReferenceType
  implements AddressingHeaderItem {

  /**
   * <code>serialVersionUID</code> attribute.
   */
  private static final long serialVersionUID = -1650335889022891672L;

  /**
   * Class logger.
   */
  private static final Log LOG = LogFactory.getLog(EndpointReference.class);

  /**
   * Creates a new EndpointReference instance.
   *
   * @param address Remote address
   */
  public EndpointReference(Address address) {
    this.setAddress(address);
  }

  /**
   * Constructor EndpointReference.
   *
   * @param element Source element
   * @throws MalformedURIException If an error occurs
   */
  public EndpointReference(Element element) throws MalformedURIException {
    NodeList children = element.getChildNodes();
    for (int i = 0; i < children.getLength(); i++) {
      Node node = children.item(i);
      if (node instanceof Element) {
        Element child = (Element) node;

        String ns = child.getNamespaceURI();
        if (AddressingUtils.getAddressingVersion(ns) == null) {
          // skip for now - does not currently handle
          // extensibility elements
          continue;
        }

        String localName = child.getLocalName();
        if (Constants.ADDRESS.equals(localName)) {
          this.setAddress(Address.fromElement(child));
        } else if (Constants.PORT_TYPE.equals(localName)) {
          this.setPortType(PortType.fromElement(child));
        } else if (Constants.SERVICE_NAME.equals(localName)) {
          this.setServiceName(ServiceNameType.fromElement(child));
        } else if (Constants.REFERENCE_PROPERTIES.equals(localName)) {
          this.setProperties(ReferencePropertiesType.fromElement(child));
        } else if (Constants.REFERENCE_PARAMETERS.equals(localName)) {
          this.setParameters(ReferenceParametersType.fromElement(child));
        } else if (Constants.METADATA.equals(localName)) {
          this.setMetadata(MetaDataType.fromElement(child));
        }
      }
    }
    if (this.getAddress() == null) {
      throw new IllegalArgumentException("Specified element does not contain the"
        + " required 'wsa:Address' child element.");
    }
  }

  /**
   * Creates a new EndpointReference instance.
   *
   * @param endpoint Remote end-point
   */
  public EndpointReference(EndpointReferenceType endpoint) {
    super(endpoint);
  }

  /**
   * Creates a new EndpointReference instance.
   *
   * @param element SOAP element node
   * @throws MalformedURIException If an error occurs
   */
  public EndpointReference(SOAPElement element) throws MalformedURIException {
    Iterator<?> iter = element.getChildElements();
    while (iter.hasNext()) {
      Object node = iter.next();

      if (node instanceof SOAPElement) {
        SOAPElement child = (SOAPElement) node;

        String ns = child.getNamespaceURI();
        if (AddressingUtils.getAddressingVersion(ns) == null) {
          // skip for now - does not currently handle extensibility elements
          continue;
        }

        String localName = child.getLocalName();
        if (Constants.ADDRESS.equals(localName)) {
          this.setAddress(Address.fromSOAPElement(child));
        } else if (Constants.PORT_TYPE.equals(localName)) {
          this.setPortType(PortType.fromSOAPElement(child));
        } else if (Constants.SERVICE_NAME.equals(localName)) {
          this.setServiceName(ServiceNameType.fromSOAPElement(child));
        } else if (Constants.REFERENCE_PROPERTIES.equals(localName)) {
          this.setProperties(ReferencePropertiesType.fromSOAPElement(child));
        } else if (Constants.REFERENCE_PARAMETERS.equals(localName)) {
          this.setParameters(ReferenceParametersType.fromSOAPElement(child));
        } else if (Constants.METADATA.equals(localName)) {
          this.setMetadata(MetaDataType.fromSOAPElement(child));
        }
      }
    }
    if (this.getAddress() == null) {
      throw new IllegalArgumentException("Specified element does not contain the"
        + " required 'wsa:Address' child element.");
    }
  }

  /**
   * Creates a new EndpointReference instance.
   *
   * @param endpoint Target end-point
   * @throws URI.MalformedURIException If the specified URI is not valid
   */
  public EndpointReference(String endpoint) throws URI.MalformedURIException {
    this.setAddress(new Address(endpoint));
  }

  /**
   * Creates a new EndpointReference instance.
   *
   * @param uri Remote address URI
   */
  public EndpointReference(URI uri) {
    if (uri != null) {
      this.setAddress(new Address(uri));
    }
  }

  /**
   * Insert a new metadata item.
   *
   * @param data XML rep. of metadata (MessageElement)
   */
  public void addMetadata(MessageElement data) {
    if (this.metaData == null) {
      this.metaData = new MetaDataType();
    }
    this.metaData.add(data);
  }

  /**
   * Insert a new reference parameter.
   *
   * @param param XML rep of referenceParameter (MessageElement)
   */
  public void addReferenceParameter(MessageElement param) {
    this.getParameters().add(param);
  }

  /**
   * Insert a new reference property.
   *
   * @param prop XML rep of referenceParameter (MessageElement)
   */
  public void addReferenceProperty(MessageElement prop) {
    this.getProperties().add(prop);
  }

  /**
   * Method toDOM.
   *
   * @param doc Document to use for Node creation
   * @return DOM Element
   */
  public Element toDOM(Document doc) {
    return this.toDOM(doc, Constants.ENDPOINT_REFERENCE);
  }

  /**
   * Method toDOM. This will create a DOM element in the supplied document
   * containing the XML-ized EPR. A namespace declaration for WSA will be added
   * to this element, using either the default namespace (if the passed
   * elementName isn't prefixed) or a prefixed namespace (if it is of the form
   * "foo:bar").
   *
   * @param doc Document to use as Node factory
   * @param elementName Element name
   * @return DOM element
   */
  public Element toDOM(Document doc, String elementName) {
    if (doc == null) {
      try {
        doc = XMLUtils.newDocument();
      } catch (ParserConfigurationException pce) {
        LOG.error("Error creating XML document", pce);
        return null;
      }
    }

    Element elem = toDOMInternal(doc, elementName);

    // !!! Shouldn't we let the user decide where to put it?? --gdaniels
    if (doc.getDocumentElement() == null) {
      doc.appendChild(elem);
    } else {
      doc.getDocumentElement().appendChild(elem);
    }
    return elem;
  }

  /**
   * This will create a DOM element containing the XML-ized EPR.
   * <p>
   * A namespace declaration for WSA will be added
   * to this element, using either the default namespace (if the passed
   * elementName isn't prefixed) or a prefixed namespace (if it is of the form
   * "foo:bar").
   *
   * @param doc         DOM node factory
   * @param elementName Name of the root element to create
   * @return Created element
   */
  private Element toDOMInternal(Document doc, String elementName) {
    int ix = elementName.indexOf(":");
    String prefix = "";
    if (ix != -1) {
      prefix = elementName.substring(0, ix);
    }

    AddressingVersion version = AddressingUtils.getAddressingVersion();
    String ns = version.getNamespace();
    Element parent = doc.createElementNS(ns, elementName);

    // Add namespace declaration
    if (prefix.equals("")) {
      parent.setAttributeNS(org.apache.axis.Constants.NS_URI_XMLNS, "xmlns", ns);
    } else {
      parent
        .setAttributeNS(org.apache.axis.Constants.NS_URI_XMLNS, "xmlns:" + prefix, ns);
      prefix = prefix + ":"; // Now we just use this to make QNames...
    }

    append(this.getAddress(), version, parent, prefix + Constants.ADDRESS);
    append(this.getPortType(), version, parent, prefix + Constants.PORT_TYPE);
    append(this.getServiceName(), version, parent, prefix + Constants.SERVICE_NAME);
    if (version.supportsProperties()) {
      append(this.getProperties(), version, parent,
             prefix + Constants.REFERENCE_PROPERTIES);
    }
    if (version.supportsParameters()) {
      append(this.getParameters(), version, parent,
             prefix + Constants.REFERENCE_PARAMETERS);
    }
    if (version.supportsMetadata()) {
      append(this.getMetadata(), version, parent, prefix + Constants.METADATA);
    }

    MessageElement[] any = this.get_any();
    if (any != null && any.length > 0) {
      for (MessageElement element : any) {
        try {
          parent.appendChild(doc.importNode(element.getAsDOM(), true));
        } catch (Exception e) {
          LOG.debug("", e);
        }
      }
    }

    return parent;
  }

  /**
   * Appends the specified instance to the parent DOM node, taking care
   * of null values.
   *
   * @param appendable   Instance to be appended
   * @param version      WS-Addressing version to use
   * @param parent       DOM Element to append to
   * @param elementName  Name of the element to create as root node of appendable
   */
  private void append(DOMAppendable appendable, AddressingVersion version,
    Element parent, String elementName) {
    if (appendable != null) {
      appendable.append(version, parent, elementName);
    }
  }

  /**
   * {@inheritDoc}
   */
  public SOAPHeaderElement toSOAPHeaderElement(AddressingVersion version,
    SOAPEnvelope env, String actorURI) throws SOAPException {
    return this.toSOAPHeaderElement(version, env, actorURI, Constants.ENDPOINT_REFERENCE);
  }


  /**
   * {@inheritDoc}
   */
  public SOAPHeaderElement toSOAPHeaderElement(SOAPEnvelope env, String actorURI)
    throws SOAPException {
    AddressingVersion version = AddressingUtils.getAddressingVersion();
    return this.toSOAPHeaderElement(version, env, actorURI);
  }

  /**
   * Translates this instance into a SOAPHeaderElement.
   *
   * @param version   WS-Addressing version to use
   * @param env       SOAP Envelope to put the element into
   * @param actorURI  Actor URI
   * @param name      Element name
   * @return Created element
   * @throws SOAPException If an error occurs
   */
  protected SOAPHeaderElement toSOAPHeaderElement(AddressingVersion version,
    SOAPEnvelope env, String actorURI, String name) throws SOAPException {

    if (name == null) {
      throw new IllegalArgumentException("Name parameter must be non-null.");
    }
    if (env == null) {
      throw new IllegalArgumentException("Env parameter must be non-null.");
    }
    if (env instanceof org.apache.axis.message.SOAPEnvelope) {

      SOAPHeader header = env.getHeader();
      if (header == null) {
        env.addHeader();
      }

      try {
        Document doc = XMLUtils.newDocument();

        org.apache.axis.message.SOAPHeaderElement headerElem
          = new org.apache.axis.message.SOAPHeaderElement(this.toDOM(doc, name));
        headerElem.setActor(actorURI);
        ((org.apache.axis.message.SOAPEnvelope) env).addHeader(headerElem);

        return headerElem;
      } catch (Exception e) {
        throw new SOAPException("Could not create XML document", e);
      }
    } else {
      throw new SOAPException("SOAPEnvelope implementation not supported: "
        + env.getClass().getName());
    }
  }

}
