/*
 *  Copyright (c) 2008 Rodrigo Ruiz
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.apache.axis.message.addressing;

import javax.xml.namespace.QName;
import javax.xml.soap.SOAPElement;

import org.apache.axis.description.TypeDesc;
import org.apache.axis.encoding.Deserializer;
import org.apache.axis.encoding.Serializer;
import org.apache.axis.encoding.SimpleType;
import org.apache.axis.encoding.ser.SimpleDeserializer;
import org.apache.axis.encoding.ser.SimpleSerializer;
import org.apache.axis.message.addressing.util.TextExtractor;
import org.apache.axis.utils.XMLUtils;
import org.w3c.dom.Element;

/**
 * Class PortType.
 *
 * @author Davanum Srinivas
 * @version $Revision: 14 $
 */
public class PortType extends AttributedQName implements SimpleType {
  /**
   * <code>serialVersionUID</code> attribute.
   */
  private static final long serialVersionUID = -3450247052859667871L;

  /**
   * Type description.
   */
  private static final TypeDesc TYPE_DESC = new TypeDesc(PortType.class, true);

  static {
    String ns = Constants.NS_URI_ADDRESSING_DEFAULT;
    TYPE_DESC.setXmlType(new QName(ns, "PortType"));
  }

  /**
   * Return the type description object.
   *
   * @return Type type description for this class
   */
  public static TypeDesc getTypeDesc() {
    return TYPE_DESC;
  }

  /**
   * Gets a custom Deserializer.
   *
   * @param mType not used
   * @param jType Java type
   * @param xType XML type
   * @return A Deserializer instance
   */
  public static Deserializer getDeserializer(String mType, Class<?> jType, QName xType) {
    return new SimpleDeserializer(jType, xType, TYPE_DESC);
  }

  /**
   * Gets a custom Serializer.
   *
   * @param mType not used
   * @param jType Java type
   * @param xType XML type
   * @return A Serializer instance
   */
  public static Serializer getSerializer(String mType, Class<?> jType, QName xType) {
    return new SimpleSerializer(jType, xType, TYPE_DESC);
  }

  /**
   * Parses an element and extracts the port type name.
   *
   * @param element Element to parse
   * @return Port type name
   */
  public static PortType fromElement(Element element) {
    String value = TextExtractor.getText(element);
    QName qname = XMLUtils.getQNameFromString(value, element);
    return new PortType(qname);
  }

  /**
   * Parses an element and extracts the port type name.
   *
   * @param element Element to parse
   * @return Port type name
   */
  public static PortType fromSOAPElement(SOAPElement element) {
    String value = TextExtractor.getText(element);
    QName qname = TextExtractor.getQName(value, element);
    return new PortType(qname);
  }

  /**
   * Constructor PortType.
   *
   * @param qname Port name
   */
  public PortType(QName qname) {
    super(qname);
  }

  /**
   * Constructor PortType.
   *
   * @param namespace Port namespace
   * @param localName Port local name
   */
  public PortType(String namespace, String localName) {
    super(namespace, localName);
  }

  /**
   * append DOM node to parent.
   *
   * @param parent Parent element
   */
  public void append(Element parent) {
    append(parent, Constants.PORT_TYPE);
  }
}
