/*
 *  Copyright (c) 2008 Rodrigo Ruiz
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.apache.axis.message.addressing;

import javax.xml.namespace.QName;
import javax.xml.soap.Name;
import javax.xml.soap.SOAPElement;
import javax.xml.soap.SOAPEnvelope;
import javax.xml.soap.SOAPException;
import javax.xml.soap.SOAPHeader;

import org.apache.axis.message.SOAPHeaderElement;
import org.apache.axis.message.addressing.util.AddressingUtils;
import org.apache.axis.message.addressing.util.TextExtractor;
import org.apache.axis.types.URI;
import org.apache.axis.types.URI.MalformedURIException;

/**
 * Java content class for RelatesTo element declaration.
 * <p>
 * The following schema fragment specifies the expected content contained within
 * this java content object. (defined at
 * http://schemas.xmlsoap.org/ws/2004/08/addressing line 77)
 * <p>
 *
 * <pre>
 * &lt;element
 *   name=&quot;RelatesTo&quot;
 *   type=&quot;{http://schemas.xmlsoap.org/ws/2004/08/addressing}Relationship&quot;/&gt;
 * </pre>
 *
 * @author Davanum Srinivas
 * @version $Revision: 14 $
 */
public class RelatesTo extends Relationship implements AddressingHeaderItem {

  /**
   * Constructor RelatesTo.
   *
   * @param element Source element
   * @throws MalformedURIException If an error occurs
   */
  public RelatesTo(SOAPElement element) throws MalformedURIException {
    super();
    AddressingVersion version = AddressingUtils.getAddressingVersion(
      element.getNamespaceURI());

    this.setURI(new URI(TextExtractor.getText(element)));
    String value = element.getAttribute(Constants.RELATIONSHIP_TYPE);
    if (value != null && value.length() > 0) {
      if (version.isW3C()) {
        this.setTypeIRI(value);
      } else {
        int pos = value.indexOf(':');
        if (pos == -1) {
          this.setType(new QName(value));
        } else {
          String namespace = element.getNamespaceURI(value.substring(0, pos));
          this.setType(new QName(namespace, value.substring(pos + 1)));
        }
      }
    } else {
      this.setType(version.getResponseRelationshipType());
    }
  }

  /**
   * Creates an instance.
   *
   * @param uri  Target URI
   * @param type URI type
   * @throws URI.MalformedURIException If the URI is not valid
   */
  public RelatesTo(String uri, QName type) throws URI.MalformedURIException {
    super(new URI(uri), type);
  }

  /**
   * Creates an instance.
   *
   * @param uri  Target URI
   * @param type Type attribute
   * @throws URI.MalformedURIException If the URI is not valid
   */
  public RelatesTo(String uri, String type) throws URI.MalformedURIException {
    super(uri, type);
  }

  /**
   * Creates an instance.
   *
   * @param uri  Target URI
   * @param type Type attribute
   */
  public RelatesTo(URI uri, QName type) {
    super(uri, type);
  }

  /**
   * {@inheritDoc}
   */
  public SOAPHeaderElement toSOAPHeaderElement(SOAPEnvelope env, String actorURI)
    throws SOAPException {
    AddressingVersion version = AddressingUtils.getAddressingVersion();
    return toSOAPHeaderElement(version, env, actorURI);
  }

  /**
   * {@inheritDoc}
   */
  public SOAPHeaderElement toSOAPHeaderElement(AddressingVersion version,
    SOAPEnvelope env, String actorURI) throws SOAPException {
    Name nm = env.createName(Constants.RELATES_TO, Constants.NS_PREFIX_ADDRESSING,
      version.getNamespace());
    SOAPHeader header = env.getHeader();
    if (header == null) {
      header = env.addHeader();
    }

    SOAPHeaderElement headerElement = (SOAPHeaderElement) header.addHeaderElement(nm);
    headerElement.setActor(actorURI);

    String typeIRI = this.getTypeIRI();
    QName type = this.getType();

    if (typeIRI != null) {
      headerElement.addAttribute("", Constants.RELATIONSHIP_TYPE, typeIRI);
    } else if (type != null) {
      headerElement.addAttribute("", Constants.RELATIONSHIP_TYPE, type);
    }

    headerElement.addTextNode(this.getURI().toString());

    return headerElement;
  }

}
