/*
 *  Copyright (c) 2008 Rodrigo Ruiz
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.apache.axis.message.addressing;

import javax.xml.namespace.QName;
import javax.xml.soap.SOAPElement;

import org.apache.axis.description.AttributeDesc;
import org.apache.axis.description.TypeDesc;
import org.apache.axis.encoding.Deserializer;
import org.apache.axis.encoding.Serializer;
import org.apache.axis.encoding.SimpleType;
import org.apache.axis.encoding.ser.SimpleDeserializer;
import org.apache.axis.encoding.ser.SimpleSerializer;
import org.apache.axis.message.addressing.util.TextExtractor;
import org.apache.axis.utils.XMLUtils;
import org.w3c.dom.Element;

/**
 * Java content class for ServiceNameType complex type.
 * <p>
 * The following schema fragment specifies the expected content contained within
 * this java content object. (defined at
 * http://schemas.xmlsoap.org/ws/2004/08/addressing line 67)
 * </p>
 *
 * <pre>
 * &lt;complexType name=&quot;ServiceNameType&quot;&gt;
 *   &lt;simpleContent&gt;
 *     &lt;extension base=&quot;&lt;http://www.w3.org/2001/XMLSchema&gt;QName&quot;&gt;
 *       &lt;attribute name=&quot;PortName&quot;
 *          type=&quot;{http://www.w3.org/2001/XMLSchema}NCName&quot; /&gt;
 *     &lt;/extension&gt;
 *   &lt;/simpleContent&gt;
 * &lt;/complexType&gt;
 * </pre>
 *
 * @author Davanum Srinivas
 * @version $Revision: 14 $
 */
public class ServiceNameType extends AttributedQName implements SimpleType {

  /**
   * <code>serialVersionUID</code> attribute.
   */
  private static final long serialVersionUID = 4494158339591159503L;

  /**
   * Type meta-data.
   */
  private static final TypeDesc TYPE_DESC = new TypeDesc(ServiceNameType.class, true);

  static {
    String ns = Constants.NS_URI_ADDRESSING_DEFAULT;
    TYPE_DESC.setXmlType(new QName(ns, "ServiceNameType"));
    AttributeDesc attrField = new AttributeDesc();
    attrField.setFieldName("port");
    attrField.setXmlName(new QName("", "PortName"));
    attrField.setXmlType(new QName("http://www.w3.org/2001/XMLSchema", "string"));
    TYPE_DESC.addFieldDesc(attrField);
  }

  /**
   * Parses an XML element node and extracts an instance of this class.
   *
   * @param element Element to parse
   * @return Resulting instance
   */
  public static ServiceNameType fromElement(Element element) {
    String value = TextExtractor.getText(element);
    QName qname = XMLUtils.getQNameFromString(value, element);
    String portName = element.getAttribute(Constants.PORT_NAME);
    portName = portName == null || portName.length() == 0 ? null : portName;
    return new ServiceNameType(qname, portName);
  }

  /**
   * Parses a SOAP element and extracts an instance of this class.
   *
   * @param element Element to parse
   * @return Resulting instance
   */
  public static ServiceNameType fromSOAPElement(SOAPElement element) {
    String value = TextExtractor.getText(element);
    QName qname = TextExtractor.getQName(value, element);
    String portName = element.getAttribute(Constants.PORT_NAME);
    portName = portName == null || portName.length() == 0 ? null : portName;
    return new ServiceNameType(qname, portName);
  }

  /**
   * Gets a custom Deserializer.
   *
   * @param mType not used
   * @param jType Java type
   * @param xType XML type
   * @return A Deserializer instance
   */
  public static Deserializer getDeserializer(String mType, Class<?> jType, QName xType) {
    return new SimpleDeserializer(jType, xType, TYPE_DESC);
  }

  /**
   * Gets a custom Serializer.
   *
   * @param mType not used
   * @param jType Java type
   * @param xType XML type
   * @return A Serializer instance
   */
  public static Serializer getSerializer(String mType, Class<?> jType, QName xType) {
    return new SimpleSerializer(jType, xType, TYPE_DESC);
  }

  /**
   * Return the type description object.
   *
   * @return Type type description for this class
   */
  public static TypeDesc getTypeDesc() {
    return TYPE_DESC;
  }

  /**
   * Port attribute.
   */
  private String port;

  /**
   * Creates a new instance.
   *
   * @param qname Service qualified name
   */
  public ServiceNameType(QName qname) {
    super(qname);
  }

  /**
   * Creates a new instance.
   *
   * @param qname Service qualified name
   * @param port Service port
   */
  public ServiceNameType(QName qname, String port) {
    super(qname);
    this.port = port;
  }

  // Axis bits to serialize/deserialize the fields correctly

  /**
   * Creates a new instance.
   *
   * @param serviceName Service name to copy
   */
  public ServiceNameType(ServiceNameType serviceName) {
    super(serviceName);
    this.port = serviceName.getPort();
  }

  /**
   * Creates a new instance.
   *
   * @param namespace Service namespace
   * @param localName Service local name
   */
  public ServiceNameType(String namespace, String localName) {
    super(namespace, localName);
  }

  /**
   * {@inheritDoc}
   */
  @Override
  public void append(AddressingVersion version, Element parent, String elementName) {
    Element sn = parent.getOwnerDocument().createElementNS(
      version.getNamespace(), elementName);
    String value = XMLUtils.getStringForQName(this, sn);
    sn.appendChild(parent.getOwnerDocument().createTextNode(value));
    if (this.getPort() != null) {
      sn.setAttribute(Constants.PORT_NAME, this.getPort());
    }
    parent.appendChild(sn);
  }

  /**
   * {@inheritDoc}
   */
  public void append(Element parent) {
    append(parent, Constants.SERVICE_NAME);
  }

  /**
   * Gets the service port.
   *
   * @return Service port name
   */
  public String getPort() {
    return this.port;
  }

  /**
   * Sets the service port.
   *
   * @param port New service port name
   */
  public void setPort(String port) {
    this.port = port;
  }

}
