/*
 *  Copyright (c) 2008 Rodrigo Ruiz
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.apache.axis.message.addressing.handler;

import org.apache.axis.AxisFault;
import org.apache.axis.Message;
import org.apache.axis.MessageContext;
import org.apache.axis.AxisEngine;
import org.apache.axis.client.Call;
import org.apache.axis.client.Service;
import org.apache.axis.message.SOAPEnvelope;
import org.apache.axis.message.SOAPBodyElement;
import org.apache.axis.message.RPCElement;
import org.apache.axis.message.SOAPFault;
import org.apache.axis.description.OperationDesc;
import org.apache.axis.components.uuid.UUIDGen;
import org.apache.axis.components.uuid.UUIDGenFactory;
import org.apache.axis.handlers.BasicHandler;
import org.apache.axis.message.addressing.Action;
import org.apache.axis.message.addressing.AddressingVersion;
import org.apache.axis.message.addressing.AttributedURI;
import org.apache.axis.message.addressing.AddressingHeaders;
import org.apache.axis.message.addressing.Constants;
import org.apache.axis.message.addressing.EndpointReference;
import org.apache.axis.message.addressing.EndpointReferenceType;
import org.apache.axis.message.addressing.MessageID;
import org.apache.axis.message.addressing.To;
import org.apache.axis.message.addressing.AddressingHeaderFault;
import org.apache.axis.message.addressing.util.AddressingUtils;
import org.apache.axis.types.URI;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import javax.xml.namespace.QName;
import java.util.ArrayList;
import java.util.List;
import java.util.StringTokenizer;

/**
 * Class AddressingHandler.
 *
 * @author Davanum Srinivas
 * @version $Revision: 14 $
 */
public class AddressingHandler extends BasicHandler implements Constants {

  /**
   * <code>serialVersionUID</code> attribute.
   */
  private static final long serialVersionUID = 831310304473973743L;

  /**
   * Class logger.
   */
  protected static final Log LOG = LogFactory.getLog(AddressingHandler.class);

  /**
   * UUID generator.
   */
  private static UUIDGen uuidGen = UUIDGenFactory.getUUIDGen();

  /**
   * A list of QNames of reference properties specified by the
   * deployer that need to be handled.
   * <p>
   * If the ignore flag (above) is set to false, and this list is
   * null - we process all headers
   */
  private List<QName> refPropQNames;

  /**
   * Actor URI.
   */
  private String actor;

  /**
   * Whether to remove headers once processed.
   */
  private boolean removeHeaders = false;

  /**
   * Whether to allow relative Action URIs or not.
   */
  private boolean allowNonSpecificActions = false;

  /**
   * This is defined in newer Axis builds in Handler.java, but replicated
   * here for backwards compatibility.  Note that some faults will NOT be
   * correctly processed using older (pre-Feb-2006) versions of Axis.
   */
  private static final String OPT_CATCH_OWN_FAULTS = "handler.catchOwnFaults";

  /**
   * Creates a new instance.
   */
  public AddressingHandler() {
  }

  /**
   * {@inheritDoc}
   */
  @Override
  public void init() {
    super.init();
    initializeReferencePropertyNames();
    initializeRemoveHeaders();
    initializeActor();
    initializeAllowNonSpecificActions();

    // Default to calling onFault() for our own faults
    Boolean catchOwnFaults = (Boolean)getOption(OPT_CATCH_OWN_FAULTS);
    if (catchOwnFaults == null) {
      setOption(OPT_CATCH_OWN_FAULTS, Boolean.TRUE);
    }
  }

  /**
   * Initializes the "actor" URI field.
   */
  protected void initializeActor() {
    actor = (String) getOption("actor");
  }

  /**
   * Initializes the "removeHeaders" configuration flag.
   */
  private void initializeRemoveHeaders() {
    String property = (String) getOption("removeHeaders");
    this.removeHeaders = "true".equalsIgnoreCase(property);
  }

  /**
   * Initializes the "allowNonSpecificActions" configuration flag.
   */
  private void initializeAllowNonSpecificActions() {
    String property = (String) getOption("allowNonSpecificActions");
    this.allowNonSpecificActions = "true".equalsIgnoreCase(property);
  }

  /**
   * {@inheritDoc}
   */
  public void invoke(MessageContext msgContext) throws AxisFault {
    boolean setMustUnderstand =
      msgContext.isPropertyTrue(ENV_ADDRESSING_SET_MUST_UNDERSTAND);

    try {
      if (msgContext.isClient()) {
        if (!msgContext.getPastPivot()) {
          processClientRequest(msgContext, setMustUnderstand);
        } else {
          processClientResponse(msgContext);
        }
      } else {
        if (!msgContext.getPastPivot()) {
          processServerRequest(msgContext);
        } else {
          processServerResponse(msgContext, setMustUnderstand);
        }
      }
    } catch (Exception e) {
      LOG.error("Exception in AddressingHandler", e);
      throw AxisFault.makeFault(e);
    }
  }

  /**
   * {@inheritDoc}
   */
  public void onFault(MessageContext msgContext) {
    if (msgContext.isClient()) {
      return;
    }
    try {
      processFault(msgContext);
    } catch (Exception e) {
      LOG.error("Exception in AddressingHandler", e);
    }
  }

  /**
   * Processes a fault.
   *
   * @param ctx Context information
   * @throws Exception If an error occurs
   */
  protected void processFault(MessageContext ctx) throws Exception {
    Message msg = ctx.getResponseMessage();
    if (msg == null) {
      return;
    }

    AddressingVersion version = AddressingUtils.getAddressingVersion(ctx);

    SOAPFault el = (SOAPFault)msg.getSOAPEnvelope().getFirstBody();
    if (el.getFault().getFaultCode().equals(
      ctx.getSOAPConstants().getMustunderstandFaultQName())) {
      return;
    }

    AddressingHeaders reqHeaders =
      (AddressingHeaders) ctx.getProperty(ENV_ADDRESSING_REQUEST_HEADERS);

    if (reqHeaders == null) {
      // error?
      return;
    }

    AddressingHeaders resHeaders = AddressingUtils.getResponseHeaders(ctx);

    // set From
    EndpointReference fromEPR = resHeaders.getFrom();
    if (fromEPR == null) {
      To toURI = reqHeaders.getTo();
      if (toURI != null) {
        fromEPR = new EndpointReference(toURI);
        fromEPR.setProperties(reqHeaders.getReferenceProperties());
        resHeaders.setFrom(fromEPR);
      }
    }
    URI actionURI = new URI(version.getFaultActionURI(), allowNonSpecificActions);
    resHeaders.setAction(actionURI);

    // process RelatesTo
    MessageID msgID = reqHeaders.getMessageID();
    if (msgID != null) {
      resHeaders.addRelatesTo(msgID.toString(), version.getResponseRelationshipType());
    }

    // process MessageID
    msgID = new MessageID(new URI("uuid:" + uuidGen.nextUUID()));
    resHeaders.setMessageID(msgID);

    // now put all headers into soap env.
    EndpointReferenceType faultTo = reqHeaders.getFaultTo();
    if (faultTo == null) {
      faultTo = reqHeaders.getReplyTo();
    }

    if (faultTo != null) {
      resHeaders.setReferenceProperties(faultTo.getProperties());
      resHeaders.setReferenceParameters(faultTo.getParameters());

      // resHeaders.toEnvelope(msg.getSOAPEnvelope());

      // process FaultTo
      AttributedURI address = faultTo.getAddress();
      if (address != null) {
        String uri = address.toString();
        /*
         * Fix Sonic00027851, WS Addressing does not properly examine and
         * deliver to an anonymous FaultTo address
         */
        if (uri != null
            && !uri.equals(version.getAnonymousRoleURI())) {
          // send the msg to fault to
          if (!uri.equals(URI_NONE_W3C_CR)) {
            forwardMessage(faultTo, msg);
          }
          // Somehow make the response empty, or create a new empty
          // response
          /*
           * Fix Sonic00027691 avoid NullPointerException with
           * msgContext.setCurrentMessage(null)
           */
          ctx.setResponseMessage(null);
          ctx.setProperty("axis.DontEmitFault", Boolean.TRUE);
          return;
        }
      }
    }

    // TODO : Default to ReplyTo, and then From?

    resHeaders.toEnvelope(msg.getSOAPEnvelope());
  }

  /**
   * Method processClientRequest.
   *
   * @param ctx Context
   * @param setMustUnderstand Must-Understand attribute value
   */
  protected void processClientRequest(MessageContext ctx,  boolean setMustUnderstand)
    throws Exception {

    boolean sendDefaultFrom = ctx.isPropertyTrue(SEND_DEFAULT_FROM, true);
    boolean sendDefaultMessageID = ctx.isPropertyTrue(SEND_DEFAULT_MESSAGEID, true);
    boolean sendDefaultTo = ctx.isPropertyTrue(SEND_DEFAULT_TO, true);

    AddressingHeaders sharedHeaders = (AddressingHeaders) ctx
    .getProperty(ENV_ADDRESSING_SHARED_HEADERS);

    AddressingHeaders headers = AddressingUtils.getRequestHeaders(ctx);

    headers.setSetMustUnderstand(setMustUnderstand);

    // set MessageID
    if (headers.getMessageID() == null && sendDefaultMessageID) {
      MessageID id = new MessageID(new URI("uuid:" + uuidGen.nextUUID()));
      headers.setMessageID(id);
    }

    // set To
    To to = headers.getTo();
    if (to == null && sendDefaultTo) {
      if (sharedHeaders != null) {
        // To is always set in shared headers
        headers.setTo(sharedHeaders.getTo());
      } else {
        headers.setTo(new To(ctx.getStrProp(MessageContext.TRANS_URL)));
      }
    }

    // set Action
    String action = ctx.getSOAPActionURI();
    if (action != null) {
      URI actionURI = new URI(action, allowNonSpecificActions);
      headers.setAction(new Action(actionURI));
    } else {
      Action act = headers.getAction();
      if (act == null) {
        URI actionURI = new URI("http://apache.org/axis/defaultAction");
        headers.setAction(new Action(actionURI));
      }

      ctx.setUseSOAPAction(true);
      // Make SOAP action match
      ctx.setSOAPActionURI(headers.getAction().toString());
    }

    // set From
    if (headers.getFrom() == null && sendDefaultFrom) {
      String from = ctx.getStrProp(ENV_ADDRESSING_FROM_URI);
      if (from == null) {
        from = AddressingUtils.getAddressingVersion().getAnonymousRoleURI();
      }
      headers.setFrom(new EndpointReference(from));
    }

    // set ReplyTo
    if (ctx.isPropertyTrue(ENV_ADDRESSING_SEND_REPLYTO) && headers.getReplyTo() == null) {
      String replyTo = ctx.getStrProp(ENV_ADDRESSING_REPLYTO_URI);
      if (replyTo != null) {
        headers.setReplyTo(new EndpointReference(replyTo));
      } else {
        headers.setReplyTo(headers.getFrom());
      }
    }

    // set FaultTo
    if (headers.getFaultTo() == null) {
      String faultTo = ctx.getStrProp(ENV_ADDRESSING_FAULTTO_URI);
      if (faultTo != null) {
        headers.setFaultTo(new EndpointReference(faultTo));
      }
    }

    if (sharedHeaders != null) {
      headers.setReferenceProperties(sharedHeaders.getReferenceProperties());
    }
    Message msg = ctx.getRequestMessage();
    headers.toEnvelope(msg.getSOAPEnvelope(), this.actor);
  }

  /**
   * Method processClientResponse.
   *
   * @param ctx Context information
   * @throws Exception If an error occurs
   */
  protected void processClientResponse(MessageContext ctx) throws Exception {
    // if no response - do nothing
    Message msg = ctx.getResponseMessage();
    if (msg == null) {
      return;
    }

    AddressingHeaders headers = new AddressingHeaders(
      msg.getSOAPEnvelope(), this.actor,
      true, this.removeHeaders, false, this.allowNonSpecificActions,
      this.refPropQNames);
    ctx.setProperty(ENV_ADDRESSING_RESPONSE_HEADERS, headers);
  }

  /**
   * Method processServerRequest.
   *
   * @param ctx Context information
   * @throws Exception If an error occurs
   */
  protected void processServerRequest(MessageContext ctx) throws Exception {

    boolean isAddrRequired = ctx.isPropertyTrue(ENV_ADDRESSING_REQUIRED, true);

    Message msg = ctx.getRequestMessage();
    if (msg == null) {
      return;
    }
    AddressingHeaders headers = new AddressingHeaders(
      msg.getSOAPEnvelope(), this.actor,
      true, this.removeHeaders, false, this.allowNonSpecificActions,
      this.refPropQNames);

    AxisFault fault = headers.getAddrHeaderFault();

    if (isAddrRequired && headers.getAction() == null) {
      if (ctx.getProperty(ENV_ADDRESSING_NAMESPACE_URI) == null) {
        ctx.setProperty(ENV_ADDRESSING_NAMESPACE_URI, NS_URI_ADDRESSING_2005_08);
      }
      fault = new AddressingHeaderFault("No Action header",
          AddressingHeaderFault.HEADER_REQUIRED, "Action");
      LOG.debug("WS-A request to endpoint " + headers.getTo()
          + " is missing the required wsa:Action header.");
    }

    ctx.setProperty(ENV_ADDRESSING_REQUEST_HEADERS, headers);

    if (fault != null) {
      throw fault;
    }

    // set the target service based on To header if it hasn't already
    // been determined.  NOTE: May want to add an option to override
    // later.
    if (ctx.getService() == null && ctx.getTargetService() == null) {
      setTargetService(ctx, headers);

      if (ctx.getService() != null) {
        resetOperations(ctx);
      }
    }
  }

  /**
   * Reset the context operations.
   *
   * @param ctx Context information
   * @throws AxisFault If an error occurs
   */
  protected void resetOperations(MessageContext ctx) throws AxisFault {
    resetContextOperations(ctx);
  }

  /**
   * Reinitializes the RPCElement with right operations
   * so that things are nicely deserialized.
   *
   * @param ctx Context information
   * @throws AxisFault If an error occurs
   */
  public static void resetContextOperations(MessageContext ctx) throws AxisFault {
    Message msg = ctx.getCurrentMessage();
    if (msg == null) {
      return;
    }
    SOAPEnvelope env = msg.getSOAPEnvelope();
    if (env == null) {
      return;
    }
    SOAPBodyElement bodyElement = env.getFirstBody();
    if (bodyElement != null && bodyElement instanceof RPCElement) {
      RPCElement element = (RPCElement) bodyElement;
      // update the operations in RPCElement
      element.updateOperationsByQName();
      OperationDesc[] operations = element.getOperations();
      // and set operation if appropriate
      if (operations == null) {
        element.updateOperationsByName();
      } else if (operations.length == 1) {
        ctx.setOperation(operations[0]);
      }
    } else {
      msg.getSOAPPartAsString();
    }
  }

  /**
   * Can be overridden by subclasses to customize
   * how the wsa:to header is interpreted.
   *
   * @param ctx     Context information
   * @param headers Header container
   */
  protected void setTargetService(MessageContext ctx, AddressingHeaders headers)
    throws Exception {
    To toURI = headers.getTo();
    if (toURI == null) {
      return;
    }
    String to = toURI.getPath();
    if (to == null) {
      return;
    }
    // set the target service
    int i = to.lastIndexOf('/');
    ctx.setTargetService(to.substring(i + 1));
  }

  /**
   * Method processServerResponse.
   *
   * @param ctx Context information
   * @param setMustUnderstand Whether to set the attribute
   * @throws Exception If an error occurs
   */
  protected void processServerResponse(MessageContext ctx,
    boolean setMustUnderstand) throws Exception {
    // if no response - do nothing
    Message msg = ctx.getResponseMessage();
    if (msg == null) {
      return;
    }

    AddressingVersion version = AddressingUtils.getAddressingVersion();

    AddressingHeaders reqHeaders = (AddressingHeaders) ctx
    .getProperty(ENV_ADDRESSING_REQUEST_HEADERS);

    if (reqHeaders == null) {
      // error?
      return;
    }

    AddressingHeaders resHeaders = AddressingUtils.getResponseHeaders(ctx);

    resHeaders.setSetMustUnderstand(setMustUnderstand);

    // set From
    EndpointReference fromEPR = resHeaders.getFrom();
    if (fromEPR == null) {
      To toURI = reqHeaders.getTo();
      if (toURI != null) {
        fromEPR = new EndpointReference(toURI);
        fromEPR.setProperties(reqHeaders.getReferenceProperties());
        resHeaders.setFrom(fromEPR);
      }
    }

    // set Action
    Action action = resHeaders.getAction();
    if (action == null) {
      String respAction = ctx.getStrProp(ENV_ADDRESSING_RESP_ACTION);
      if (respAction == null) {
        // not set - try request headers
        action = reqHeaders.getAction();
        if (action != null) {
          URI actionURI = new URI(action.toString() + "Response",
              allowNonSpecificActions);
          resHeaders.setAction(new Action(actionURI));
        }
      } else {
        resHeaders.setAction(new Action(new URI(respAction)));
      }
    }

    if (resHeaders.getFrom() == null && reqHeaders.getFrom() != null) {
      resHeaders.setTo(reqHeaders.getFrom().getAddress());
    } else {
      resHeaders.setTo(new To(version.getAnonymousRoleURI()));
    }

    // process RelatesTo
    MessageID msgID = reqHeaders.getMessageID();
    if (msgID != null) {
      resHeaders.addRelatesTo(msgID.toString(), version.getResponseRelationshipType());
    }

    // process MessageID
    msgID = new MessageID(new URI("uuid:" + uuidGen.nextUUID()));
    resHeaders.setMessageID(msgID);

    // process ReplyTo
    EndpointReferenceType replyTo = reqHeaders.getReplyTo();
    if (replyTo != null) {
      resHeaders.setReferenceProperties(replyTo.getProperties());
      resHeaders.setReferenceParameters(replyTo.getParameters());

      AttributedURI address = replyTo.getAddress();
      if (address != null) {
        String uri = address.toString();
        if (uri != null && !uri.equals(version.getAnonymousRoleURI())) {
          if (!uri.equals(URI_NONE_W3C_CR)) {
            // send the msg to reply to
            forwardMessage(replyTo, msg);
          }
          // Nothing goes back out the HTTP connection.
          ctx.setResponseMessage(null);
        } else {
          resHeaders.toEnvelope(msg.getSOAPEnvelope(), this.actor);
        }
      }
    } else {
      resHeaders.toEnvelope(msg.getSOAPEnvelope(), this.actor);
    }
  }

  /**
   * Forwards the message to the given EPR.
   *
   * @param epr Destination EPR
   * @param msg Message to forward
   * @throws Exception If an error occurs
   */
  protected void forwardMessage(EndpointReferenceType epr, Message msg)
    throws Exception {
    AttributedURI address = epr.getAddress();

    AddressingHeaders headers = null;
    MessageContext ctx = msg.getMessageContext();
    if (ctx != null) {
      headers = (AddressingHeaders) ctx
      .getProperty(ENV_ADDRESSING_RESPONSE_HEADERS);
    }
    if (headers == null) {
      headers = new AddressingHeaders();
      headers.setTo(address);
      headers.setReferenceProperties(epr.getProperties());
      headers.setReferenceParameters(epr.getParameters());
    }

    Service service = getService(ctx);

    Call c = (Call) service.createCall();
    c.setSOAPVersion((ctx == null)
      ? org.apache.axis.Constants.DEFAULT_SOAP_VERSION
      : ctx.getSOAPConstants());
    c.setTargetEndpointAddress(address.toString());
    c.setRequestMessage(msg);
    c.setProperty(ENV_ADDRESSING_REQUEST_HEADERS, headers);
    c.setProperty(ENV_ADDRESSING_NAMESPACE_URI, AddressingUtils
        .getAddressingNamespaceURI(ctx));
    if (ctx != null) {
      c.setProperty(AxisEngine.PROP_SEND_XSI, ctx
          .getProperty(AxisEngine.PROP_SEND_XSI));
      c.setOperationStyle(ctx.getOperationStyle());
      c.setOperationUse(ctx.getOperationUse());
    }
    configureCall(c, ctx);
    try {
      c.invoke();
    } catch (AxisFault axisFault) {
      // This fault shouldn't necessarily become the response of
      // the original interaction, should it?  Should we be doing
      // this on a separate thread?
      LOG.error(axisFault.dumpToString());
    }
  }

  /**
   * Override this method if you need something other than the default Service.<br>
   * The service returned by this method is used in creating the new Call object.
   * Something like: <br>
   * <pre>
   * Service service = getService(msgContext);
   * Call call = service.createCall()
   * call.setTargetEndpointAddress(toEndPointReference.getAddress().toString());
   * </pre>
   *
   * @param ctx Context information
   * @return A service instance
   */
  protected Service getService(MessageContext ctx) {
    return new Service();
  }

  /**
   * Override this method to prepare the new call, for instance to add
   * properties from the old MessageContext that may be needed by other
   * handlers.
   *
   * @param call Call object about to be invoked
   * @param oldContext MessageContext of the original request/response.
   */
  protected void configureCall(Call call, MessageContext oldContext) {
  }

  /**
   * Retrieve QNames for reference properties from deployment
   * and initialize the refPropQNames list which is the list
   * of reference properties that this handler should care
   * about.
   */
  private void initializeReferencePropertyNames() {
    // check if the user wants to process all headers, this means
    // the user wants to treat non ws-addr headers as reference props
    // If this option is not enabled, the non ws-addr headers will not
    // be deserialized.
    String refPropNames = (String) getOption("referencePropertyNames");
    if (refPropNames == null) {
      this.refPropQNames = new ArrayList<QName>();
    } else if (refPropNames.equals("*")) {
      this.refPropQNames = null;
    } else {
      this.refPropQNames = new ArrayList<QName>();
      StringTokenizer tkn = new StringTokenizer(refPropNames, ",");
      while (tkn.hasMoreTokens()) {
        String qnameString = tkn.nextToken().trim();
        try {
          QName qname = QName.valueOf(qnameString);
          refPropQNames.add(qname);
        } catch (Exception e) {
          // Ignore QNames which were written incorrectly
          // if their parsing results in errors
          LOG.debug("Invalid QName ignored: " + qnameString);
        }
      }
    }
  }

}
