/*
 *  Copyright (c) 2008 Rodrigo Ruiz
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.apache.axis.message.addressing.tools.wsdl;

import java.io.IOException;
import java.io.PrintWriter;
import java.util.Map;

import javax.wsdl.Binding;
import javax.wsdl.Port;
import javax.wsdl.Service;

import org.apache.axis.utils.JavaUtils;
import org.apache.axis.utils.Messages;
import org.apache.axis.wsdl.symbolTable.BindingEntry;
import org.apache.axis.wsdl.symbolTable.PortTypeEntry;
import org.apache.axis.wsdl.symbolTable.ServiceEntry;
import org.apache.axis.wsdl.symbolTable.SymbolTable;
import org.apache.axis.wsdl.toJava.Emitter;
import org.apache.axis.wsdl.toJava.JavaBindingWriter;
import org.apache.axis.wsdl.toJava.Utils;

/**
 *
 * @author Jarek Gawor
 * @author Rodrigo Ruiz
 * @version $Revision: 14 $
 */
public class JavaAddressingServiceIfaceImplWriter extends JavaClassWithImportsWriter {

  /**
   * Creates an instance.
   *
   * @param emitter      Emitter
   * @param sEntry       Service
   * @param symbolTable  Symbol table
   */
  protected JavaAddressingServiceIfaceImplWriter(Emitter emitter, ServiceEntry sEntry,
      SymbolTable symbolTable) {
    super(emitter, sEntry, symbolTable, sEntry.getName() + "Addressing", "service");
  }

  /**
   * {@inheritDoc}
   */
  protected String getExtendsText() {
    return "extends " + resolve(this.sEntry.getName());
  }

  /**
   * {@inheritDoc}
   */
  protected String getClassText() {
    return "interface ";
  }

  /**
   * {@inheritDoc}
   */
  protected String getImplementsText() {
    return null;
  }

  /**
   * Write the body of the service file.
   *
   * @param pw Writer to print to
   * @throws IOException If an error occurs
   */
  protected void writeFileBody(PrintWriter pw) throws IOException {
    Service service = sEntry.getService();

    // get ports
    Map<?, ?> portMap = service.getPorts();

    // write a get method for each of the ports with a SOAP binding
    for (Object item : portMap.values()) {
      Port p = (Port)item;
      Binding binding = p.getBinding();

      if (binding == null) {
        throw new IOException(Messages.getMessage("emitFailNoBinding01", new String[] { p
            .getName() }));
      }

      BindingEntry bEntry = symbolTable.getBindingEntry(binding.getQName());

      if (bEntry == null) {
        throw new IOException(Messages.getMessage("emitFailNoBindingEntry01",
            new String[] { binding.getQName().toString() }));
      }

      PortTypeEntry ptEntry = symbolTable.getPortTypeEntry(binding.getPortType()
          .getQName());

      if (ptEntry == null) {
        throw new IOException(Messages.getMessage("emitFailNoPortType01",
            new String[] { binding.getPortType().getQName().toString() }));
      }

      // If this isn't an SOAP binding, skip it
      if (bEntry.getBindingType() != BindingEntry.TYPE_SOAP) {
        continue;
      }

      String portName = p.getName();

      if (!JavaUtils.isJavaId(portName)) {
        portName = Utils.xmlNameToJavaClass(portName);
      }

      String bindingType = (String) bEntry
      .getDynamicVar(JavaBindingWriter.INTERFACE_NAME);

      pw.println("  /**");
      pw.println("   * Gets a Port instance from an end-point reference.");
      pw.println("   *");
      pw.println("   * @param reference EPR instance");
      pw.println("   * @return Port instance");
      pw.println("   * @throws ServiceException If an error occurs");
      pw.println("   */");
      pw.println("  " + resolve(bindingType) + " get" + portName
          + "(EndpointReferenceType reference)");
      pw.println("    throws ServiceException;");
      pw.println();
    }

    pw.println();
  }

  /**
   * {@inheritDoc}
   */
  @Override
  protected void writeClassComment(PrintWriter pw) throws IOException {
    pw.println("/**");
    pw.println(" * Service interface with addressing support.");
    pw.println(" */");
  }

}
