/*
 *  Copyright (c) 2008 Rodrigo Ruiz
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.apache.axis.message.addressing.tools.wsdl;

import java.io.IOException;
import java.io.PrintWriter;
import java.util.Map;

import javax.wsdl.Binding;
import javax.wsdl.Port;
import javax.wsdl.Service;

import org.apache.axis.utils.JavaUtils;
import org.apache.axis.utils.Messages;
import org.apache.axis.wsdl.symbolTable.BindingEntry;
import org.apache.axis.wsdl.symbolTable.PortTypeEntry;
import org.apache.axis.wsdl.symbolTable.ServiceEntry;
import org.apache.axis.wsdl.symbolTable.SymbolTable;
import org.apache.axis.wsdl.toJava.Emitter;
import org.apache.axis.wsdl.toJava.JavaBindingWriter;
import org.apache.axis.wsdl.toJava.Utils;

/**
 *
 * @author Jarek Gawor
 * @author Rodrigo Ruiz
 * @version $Revision: 14 $
 */
public class JavaAddressingServiceImplWriter extends JavaClassWithImportsWriter {

  /**
   * Creates an instance.
   *
   * @param emitter      Emitter
   * @param sEntry       Service
   * @param symbolTable  Symbol table
   */
  protected JavaAddressingServiceImplWriter(Emitter emitter, ServiceEntry sEntry,
    SymbolTable symbolTable) {
    super(emitter, sEntry, symbolTable, sEntry.getName() + "AddressingLocator",
      "service");
  }

  /**
   * {@inheritDoc}
   */
  @Override protected void initImports() {
    super.initImports();
    addImport("org.apache.axis.message.addressing.AddressingHeaders");
    addImport("org.apache.axis.message.addressing.AttributedURI");
    addImport("org.apache.axis.message.addressing.Constants");
    addImport("org.apache.axis.EngineConfiguration");
    addImport("javax.xml.namespace.QName");
    addImport("java.net.URL");
    addImport("java.net.MalformedURLException");
  }

  /**
   * {@inheritDoc}
   */
  protected String getExtendsText() {
    return "extends " + resolve(this.sEntry.getName() + "Locator");
  }

  /**
   * {@inheritDoc}
   */
  protected String getImplementsText() {
    return "implements " + resolve(this.sEntry.getName() + "Addressing");
  }

  /**
   * Write the body of the service file.
   *
   * @param pw Writer to print to
   * @throws IOException If an error occurs
   */
  protected void writeFileBody(PrintWriter pw) throws IOException {
    Service service = sEntry.getService();

    writeConstructors(pw);

    // get ports
    Map<?, ?> portMap = service.getPorts();

    // write a get method for each of the ports with a SOAP binding
    for (Object item : portMap.values()) {
      Port p = (Port)item;
      Binding binding = p.getBinding();

      if (binding == null) {
        throw new IOException(Messages.getMessage("emitFailNoBinding01", new String[] { p
            .getName() }));
      }

      BindingEntry bEntry = symbolTable.getBindingEntry(binding.getQName());

      if (bEntry == null) {
        throw new IOException(Messages.getMessage("emitFailNoBindingEntry01",
            new String[] { binding.getQName().toString() }));
      }

      PortTypeEntry ptEntry = symbolTable.getPortTypeEntry(binding.getPortType()
          .getQName());

      if (ptEntry == null) {
        throw new IOException(Messages.getMessage("emitFailNoPortType01",
            new String[] { binding.getPortType().getQName().toString() }));
      }

      // If this isn't an SOAP binding, skip it
      if (bEntry.getBindingType() != BindingEntry.TYPE_SOAP) {
        continue;
      }

      String portName = p.getName();

      if (!JavaUtils.isJavaId(portName)) {
        portName = Utils.xmlNameToJavaClass(portName);
      }

      String bindingType = (String) bEntry
      .getDynamicVar(JavaBindingWriter.INTERFACE_NAME);

      pw.println("  /**");
      pw.println("   * {@inheritDoc}");
      pw.println("   */");
      pw.println("  public " + resolve(bindingType) + " get" + portName
          + "(EndpointReferenceType reference)");
      pw.println("    throws ServiceException {");
      pw.println("    AttributedURI address = reference.getAddress();");

      pw.println("    if (address == null) {");
      pw.println("      throw new ServiceException("
        + "\"No address in EndpointReference\");");
      pw.println("    }");
      pw.println("    URL endpoint;");
      pw.println("    try {");
      pw.println("      endpoint = new URL(address.toString());");
      pw.println("    } catch (MalformedURLException e) {");
      pw.println("      throw new ServiceException(e);");
      pw.println("    }");
      pw.println("    " + resolve(bindingType) + " stub = " + "get" + portName
                 + "(endpoint);");
      pw.println("    if (stub != null) {");
      pw.println("      AddressingHeaders headers = new AddressingHeaders();");
      pw.println("      headers.setTo(address);");
      pw.println("      headers.setReferenceProperties(reference.getProperties());");
      pw.println("      ((javax.xml.rpc.Stub)stub)._setProperty(");
      pw.println("        Constants.ENV_ADDRESSING_SHARED_HEADERS, headers);");
      pw.println("    }");
      pw.println("    return stub;");
      pw.println("  }");
      pw.println();
    }

    pw.println();
  }

  /**
   * Writes the constructors of the generated class.
   *
   * @param pw Writer to print to
   * @throws IOException If an error occurs
   */
  protected void writeConstructors(PrintWriter pw) throws IOException {

    String simpleClassName = resolve(this.sEntry.getName()) + "AddressingLocator";

    // Generate a default constructor
    pw.println("  /**");
    pw.println("   * Creates an instance.");
    pw.println("   */");
    pw.println("  public " + simpleClassName + "() {");
    pw.println("  }");
    pw.println();

    // Generate a constructor with a custom configuration
    pw.println("  /**");
    pw.println("   * Creates an instance.");
    pw.println("   *");
    pw.println("   * @param config Engine configuration to use");
    pw.println("   */");
    pw.println("  public " + simpleClassName + "(EngineConfiguration config) {");
    pw.println("    super(config);");
    pw.println("  }");
    pw.println();

    // Generate a constructor with WSDL and service name parameters
    pw.println("  /**");
    pw.println("   * Creates an instance.");
    pw.println("   *");
    pw.println("   * @param wsdlLoc WSDL path");
    pw.println("   * @param sName   Service name");
    pw.println("   * @throws ServiceException If an error occurs");
    pw.println("   */");
    pw.println("  public " + simpleClassName + "(String wsdlLoc, QName sName)");
    pw.println("    throws ServiceException {");
    pw.println("    super(wsdlLoc, sName);");
    pw.println("  }");
    pw.println();
  }

  /**
   * {@inheritDoc}
   */
  @Override
  protected void writeClassComment(PrintWriter pw) throws IOException {
    pw.println("/**");
    pw.println(" * Service interface with addressing support implementation.");
    pw.println(" */");
  }

}
