/*
 *  Copyright (c) 2008 Rodrigo Ruiz
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.apache.axis.message.addressing.util;

import java.io.StringReader;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.wsdl.Input;
import javax.wsdl.Operation;
import javax.wsdl.Output;
import javax.wsdl.extensions.AttributeExtensible;
import javax.xml.namespace.QName;
import javax.xml.rpc.handler.MessageContext;
import javax.xml.soap.Name;
import javax.xml.soap.SOAPElement;
import javax.xml.soap.SOAPException;
import javax.xml.soap.SOAPHeader;
import javax.xml.soap.SOAPHeaderElement;

import org.apache.axis.AxisEngine;
import org.apache.axis.message.addressing.AddressingHeaders;
import org.apache.axis.message.addressing.AddressingVersion;
import org.apache.axis.message.addressing.Constants;
import org.apache.axis.message.addressing.EndpointReference;
import org.apache.axis.utils.DOM2Writer;
import org.apache.axis.utils.XMLUtils;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.InputSource;

/**
 * A set of WS-Addressing-related utility methods.
 *
 * @author Jarek Gawor
 * @author Rodrigo Ruiz
 * @version $Revision: 14 $
 */
public abstract class AddressingUtils implements Constants {

  /**
   * A set that contains the namespace URIs (as Strings) for the various
   * versions of WSA that we support.
   */
  private static final Set<String> ADDRESSING_NS_URI_SET = new HashSet<String>();

  /**
   * A set that contains the name-space URIs (as Strings) for the various
   * versions of W3C WS-Addressing supported by this library.
   */
  private static final Set<String> W3C_ADDRESSING_NS_URI_SET = new HashSet<String>();

  /**
   * Maps name-spaces to AddressingVersion instances.
   */
  private static final Map<String, AddressingVersion> VERSIONS
    = new HashMap<String, AddressingVersion>();

  /**
   * Default addressing namespace. This namespace is used only if none is
   * configured at MessageContext level.
   */
  private static final String DEFAULT_ADDRESSING_NS;

  static {
    addVersion(NS_URI_ADDRESSING_2003_03, null, RESPONSE, false, false, true, false);
    addVersion(NS_URI_ADDRESSING_2004_03, null, REPLY, false, false, true, false);
    addVersion(NS_URI_ADDRESSING_2004_08, null, REPLY, false, true, true, false);

    addVersion(NS_URI_ADDRESSING_2005_08, URI_ANONYMOUS_W3C_CR, REPLY,
               true, true, true, true);

    // Determine default addressing namespace once
    DEFAULT_ADDRESSING_NS = System.getProperty(
      Constants.ENV_ADDRESSING_NAMESPACE_URI,
      Constants.NS_URI_ADDRESSING_DEFAULT);
  }

  /**
   * Registers a WS-Addressing version.
   *
   * @param ns       Namespace
   * @param anonURI  Anonymous target URI
   * @param response Response/Reply local name
   * @param w3c      W3C compliant?
   * @param params   Supports reference parameters?
   * @param props    Supports reference properties?
   * @param mdata    Supports meta-data?
   */
  private static void addVersion(String ns, String anonURI, String response,
    boolean w3c, boolean params, boolean props, boolean mdata) {
    if (w3c) {
      W3C_ADDRESSING_NS_URI_SET.add(ns);
    } else {
      ADDRESSING_NS_URI_SET.add(ns);
    }

    if (anonURI == null) {
      anonURI = ns + "/role/anonymous";
    }
    VERSIONS.put(ns, new WsaVersion(ns, anonURI, new QName(ns, response),
                 w3c, params, props, mdata));
  }

  /**
   * Hidden constructor.
   */
  protected AddressingUtils() {
  }

  /**
   * Removes all instances of a given header name in WSA namespace for a given
   * actor.
   *
   * @param soapHeader Header containing the element to remove
   * @param actorURI   Actor URI
   * @param headerName element name
   */
  public static void removeHeader(SOAPHeader soapHeader, String actorURI,
    String headerName) {
    if (soapHeader == null) {
      return;
    }
    Iterator<?> headers = soapHeader.examineHeaderElements(actorURI);
    List<SOAPHeaderElement> existingElements = new ArrayList<SOAPHeaderElement>();
    while (headers.hasNext()) {
      SOAPHeaderElement hElement = (SOAPHeaderElement) headers.next();
      Name hName = hElement.getElementName();
      if (isAddressingNamespaceURI(hName.getURI())
        && hName.getLocalName().equals(headerName)) {
        existingElements.add(hElement);
      }
    }
    for (int i = 0; i < existingElements.size(); i++) {
      SOAPHeaderElement el = (SOAPHeaderElement) existingElements.get(i);
      el.detachNode();
    }
  }

  /**
   * Gets an object describing the WS-Addressing specification for which the
   * specified message context is configured.
   *
   * @param ctx Message context
   * @return An AddressingVersion instance, or null
   */
  public static AddressingVersion getAddressingVersion(MessageContext ctx) {
    return VERSIONS.get(getAddressingNamespaceURI(ctx));
  }

  /**
   * Gets an object describing the WS-Addressing specification represented
   * by the specified namespace.
   *
   * @param ns Addressing namespace URI
   * @return An AddressingVersion instance, or null
   */
  public static AddressingVersion getAddressingVersion(String ns) {
    return VERSIONS.get(ns);
  }

  /**
   * Gets an object describing the WS-Addressing specification represented
   * by the namespace configured at the MessageContext or System level.
   *
   * @return An AddressingVersion instance, or null
   */
  public static AddressingVersion getAddressingVersion() {
    String ns = getAddressingNamespaceURI(AxisEngine.getCurrentMessageContext());
    return VERSIONS.get(ns);
  }

  /**
   * Returns true if the specified URI is corresponds to the namespace URI of a
   * supported version of the WSA specification, or false otherwise.
   *
   * @param uri a URI
   * @return true if the specified URI is a valid addressing namespace
   * @deprecated
   */
  public static boolean isAddressingNamespaceURI(String uri) {
    return ADDRESSING_NS_URI_SET.contains(uri);
  }

  /**
   * Returns true if the specified URI is corresponds to the namespace URI of a
   * supported version of the WSA specification using the W3C specs, or false
   * otherwise.
   *
   * @param uri a URI
   * @return true if the specified URI is a valid W3C addressing namespace
   * @deprecated
   */
  public static boolean isW3CAddressingNamespaceURI(String uri) {
    return W3C_ADDRESSING_NS_URI_SET.contains(uri);
  }

  /**
   * Removes all instances of all WSA headers for a given actor.
   *
   * @param soapHeader header container
   * @param actorURI   Actor URI
   */
  public static void removeHeaders(SOAPHeader soapHeader, String actorURI) {
    if (soapHeader == null) {
      return;
    }
    Iterator<?> headers = soapHeader.examineHeaderElements(actorURI);
    List<SOAPHeaderElement> existingElements = new ArrayList<SOAPHeaderElement>();
    while (headers.hasNext()) {
      SOAPHeaderElement hElement = (SOAPHeaderElement) headers.next();
      Name hName = hElement.getElementName();
      AddressingVersion version = getAddressingVersion(hName.getURI());
      if (version != null) {
        existingElements.add(hElement);
      }
    }
    for (int i = 0; i < existingElements.size(); i++) {
      SOAPHeaderElement el = (SOAPHeaderElement) existingElements.get(i);
      el.detachNode();
    }
  }

  /**
   * Gets response <code>AddressingHeaders</code> associated with the
   * specified context. If there are no <code>AddressingHeaders</code>
   * associated with the context, new reponse <code>AddressingHeaders</code>
   * are created and associated with the context.
   *
   * @param msgCtx Context information
   * @return The addressing headers container
   */
  public static AddressingHeaders getResponseHeaders(MessageContext msgCtx) {
    return getHeaders(msgCtx, Constants.ENV_ADDRESSING_RESPONSE_HEADERS);
  }

  /**
   * Gets request <code>AddressingHeaders</code> associated with the specified
   * context. If there are no <code>AddressingHeaders</code> associated with
   * the context, new request <code>AddressingHeaders</code> are created and
   * associated with the context.
   *
   * @param msgCtx Context information
   * @return The headers container
   */
  public static AddressingHeaders getRequestHeaders(MessageContext msgCtx) {
    return getHeaders(msgCtx, Constants.ENV_ADDRESSING_REQUEST_HEADERS);
  }

  /**
   * Gets input action. See section 3.3 of the 2004/08 WSA specification.
   *
   * @param portTypeQName Name of the port type
   * @param operation     Operation in the specified port type
   * @return The operation input action URI
   */
  public static String getInputAction(QName portTypeQName, Operation operation) {
    if (portTypeQName == null || operation == null) {
      throw new IllegalArgumentException("Null parameters are not allowed.");
    }
    Input input = operation.getInput();
    if (input == null) {
      return null;
    }
    return getAction(input, input.getName(), portTypeQName, operation, "Request");
  }

  /**
   * Gets output action. See section 3.3 of the 2004/08 WSA specification.
   *
   * @param portTypeQName Name of the port type
   * @param operation     Operation in the specified port type
   * @return The operation output action URI
   */
  public static String getOutputAction(QName portTypeQName, Operation operation) {
    if (portTypeQName == null || operation == null) {
      throw new IllegalArgumentException("Null parameters are not allowed.");
    }
    Output output = operation.getOutput();
    if (output == null) {
      return null;
    }
    return getAction(output, output.getName(), portTypeQName, operation, "Response");
  }

  /**
   * Gets prefix for a given namespace. If the prefix is already defined for the
   * namespace given, it is returned. Otherwise, a new prefix is automatically
   * registered with the given namespace and returned.
   *
   * @param element   Element where the prefix will be look up
   * @param namespace Namespace to look up
   * @return A namespace prefix
   */
  public static String getNamespacePrefix(SOAPElement element, String namespace)
    throws SOAPException {
    if (element == null || namespace == null) {
      throw new IllegalArgumentException("Null parameters are not allowed.");
    }
    Iterator<?> iter = element.getVisibleNamespacePrefixes();
    String prefix;
    String ns;
    while (iter.hasNext()) {
      prefix = (String) iter.next();
      ns = element.getNamespaceURI(prefix);
      if (namespace.equals(ns)) {
        return prefix;
      }
    }

    int i = 0;
    prefix = "ns" + i;
    while (element.getNamespaceURI(prefix) != null) {
      prefix = "ns" + i++;
    }

    element.addNamespaceDeclaration(prefix, namespace);

    return prefix;
  }

  /**
   * Returns the namespace URI of the WSA version being used by the current
   * request.
   *
   * @return The Addressing namespace URI
   * @deprecated
   */
  public static String getAddressingNamespaceURI() {
    // NOTE: the below call is Axis-specific :-(
    MessageContext ctx = AxisEngine.getCurrentMessageContext();
    return getAddressingNamespaceURI(ctx);
  }

  /**
   * Returns the namespace URI of the WSA version being used by the request
   * in the specified message context.
   *
   * @param ctx The message context
   * @return The Addressing namespace URI
   */
  public static String getAddressingNamespaceURI(MessageContext ctx) {
    String nsURI = null;
    if (ctx != null) {
      nsURI = (String) ctx.getProperty(Constants.ENV_ADDRESSING_NAMESPACE_URI);
    }
    if (nsURI == null) {
      nsURI = DEFAULT_ADDRESSING_NS;
    }
    return nsURI;
  }

  /**
   * Returns the fault wsa:Action URI that is appropriate for the WSA version
   * being used.
   *
   * @return The fault Action URI
   * @deprecated
   */
  public static String getFaultActionURI() {
    return getAddressingVersion().getFaultActionURI();
  }

  /**
   * Returns the anonymous target URI that is appropriate for the WSA version
   * being used.
   *
   * @return The anonymous target URI
   * @deprecated
   */
  public static String getAnonymousRoleURI() {
    AddressingVersion version = getAddressingVersion();
    return version.getAnonymousRoleURI();
  }

  /**
   * Returns the anonymous target URI that is appropriate for the WSA version
   * being used.
   *
   * @param ctx Message Context
   * @return The anonymous target URI
   */
  public static String getAnonymousRoleURI(MessageContext ctx) {
    return getAddressingVersion(ctx).getAnonymousRoleURI();
  }

  /**
   * Returns the response/reply relationship QName that is appropriate for the
   * WSA version being used.
   *
   * @return The response/reply QName
   */
  public static QName getResponseRelationshipType() {
    return getAddressingVersion().getResponseRelationshipType();
  }

  /**
   * Returns the response/reply relationship QName that is appropriate for the
   * WSA version being used.
   *
   * @param ctx Message context
   * @return The response/reply QName
   */
  public static QName getResponseRelationshipType(MessageContext ctx) {
    return getAddressingVersion(ctx).getResponseRelationshipType();
  }


  /**
   * Parses a string containing an End-point reference in XML format, and
   * translates it into an {@link EndpointReference} bean.
   *
   * @param epr End-point reference in XML format
   * @return Bean instance
   * @throws Exception If the XML parser fails to initialize, a syntax error
   *                   is detected in the input string, or a general error
   *                   occurs while constructing the bean
   */
  public static EndpointReference parseReference(String epr) throws Exception {
    Document doc = XMLUtils.newDocument(new InputSource(new StringReader(epr)));
    Element elem = doc.getDocumentElement();
    return new EndpointReference(elem);
  }

  /**
   * Converts an End-Point Reference into an XML String.
   *
   * @param epr End-point reference
   * @param prettyPrint Whether to beautify the generated XML text
   * @return XML String
   * @throws Exception If an error occurs
   */
  public static String toString(EndpointReference epr, boolean prettyPrint)
    throws Exception {

    if (epr == null) {
      return null;
    } else {
      Document doc = XMLUtils.newDocument();
      Element elem = epr.toDOM(doc, "wsa:EndpointReference");

      StringWriter writer = new StringWriter();
      DOM2Writer.serializeAsXML(elem, writer, true, prettyPrint);
      return writer.toString();
    }
  }

  /**
   * Gets the addressing headers from the context information.
   *
   * @param ctx Context information
   * @param type       Property name
   * @return Headers container
   */
  private static AddressingHeaders getHeaders(MessageContext ctx, String type) {
    AddressingHeaders headers = (AddressingHeaders) ctx.getProperty(type);

    if (headers == null) {
      headers = new AddressingHeaders();
      headers.isW3CVersion = getAddressingVersion(ctx).isW3C();

      // set property so other handlers might have a chance to use/modify
      ctx.setProperty(type, headers);
    }

    return headers;
  }

  /**
   * Gets an Action URI.
   *
   * @param inputOutput     Attribute containing the service WSDL
   * @param inputOutputName Name of the input/output element
   * @param portTypeQName   Port type name
   * @param operation       Operation name
   * @param defaultName     prefix to add to the operation name if inputOutputName is null
   * @return An Action URI
   */
  private static String getAction(AttributeExtensible inputOutput,
    String inputOutputName, QName portTypeQName, Operation operation,
    String defaultName) {

    Object value = inputOutput.getExtensionAttribute(new QName(
      getAddressingNamespaceURI(), Constants.WSDL_ATTRIB_ACTION));
    if (value != null) {
      // wsdl4j returns a qname by default?
      if (value instanceof QName) {
        return ((QName) value).getLocalPart();
      } else {
        return value.toString();
      }
    }
    String name = inputOutputName;
    if (name == null) {
      name = operation.getName() + defaultName;
    }
    StringBuffer buf = new StringBuffer();
    buf.append(portTypeQName.getNamespaceURI());
    if (!portTypeQName.getNamespaceURI().endsWith("/")) {
      buf.append("/");
    }
    buf.append(portTypeQName.getLocalPart()).append("/").append(name);
    return buf.toString();
  }

}
