/*
 *  Copyright (c) 2008 Rodrigo Ruiz
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.apache.ws.addressing.handler;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

import javax.xml.namespace.QName;
import javax.xml.rpc.JAXRPCException;
import javax.xml.rpc.handler.GenericHandler;
import javax.xml.rpc.handler.HandlerInfo;
import javax.xml.rpc.handler.MessageContext;

import org.apache.axis.message.addressing.Constants;
import org.apache.axis.message.addressing.MessageID;
import org.apache.axis.types.URI;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.ws.addressing.uuid.UUIdGeneratorFactory;

/**
 * Abstract JAX-RPC {@link javax.xml.rpc.handler.Handler} containing base WS-Addressing
 * functionality that is used by both client-side and server-side subclasses.
 *
 * @author Davanum Srinivas
 * @author Ian P. Springer
 * @version $Revision: 14 $
 */
public abstract class AbstractAddressingHandler extends GenericHandler {

  /**
   * Configuration property name for the Actor URI.
   */
  public static final String CONFIG_PROP_ACTOR = "actor";

  /**
   * Configuration property name for a list of reference property names.
   */
  public static final String CONFIG_PROP_REFERENCE_PROPERTY_NAMES
    = "referencePropertyNames";

  /**
   * Configuration property name for the "processed header removal" flag.
   */
  public static final String CONFIG_PROP_REMOVE_HEADERS = "removeHeaders";

  /**
   * URI scheme used in UUIDs.
   */
  static final String UUID_URI_SCHEME = "uuid";

  /**
   * List of understood addressing headers.
   */
  private static final QName[] UNDERSTOOD_HEADERS = new QName[] {
    new QName(Constants.NS_URI_ADDRESSING_2003_03, Constants.ACTION),
    new QName(Constants.NS_URI_ADDRESSING_2003_03, Constants.FAULT_TO),
    new QName(Constants.NS_URI_ADDRESSING_2003_03, Constants.FROM),
    new QName(Constants.NS_URI_ADDRESSING_2003_03, Constants.MESSAGE_ID),
    new QName(Constants.NS_URI_ADDRESSING_2003_03, Constants.RECIPIENT),
    new QName(Constants.NS_URI_ADDRESSING_2003_03, Constants.RELATES_TO),
    new QName(Constants.NS_URI_ADDRESSING_2003_03, Constants.REPLY_TO),
    new QName(Constants.NS_URI_ADDRESSING_2003_03, Constants.TO),
    new QName(Constants.NS_URI_ADDRESSING_2004_03, Constants.ACTION),
    new QName(Constants.NS_URI_ADDRESSING_2004_03, Constants.FAULT_TO),
    new QName(Constants.NS_URI_ADDRESSING_2004_03, Constants.FROM),
    new QName(Constants.NS_URI_ADDRESSING_2004_03, Constants.MESSAGE_ID),
    new QName(Constants.NS_URI_ADDRESSING_2004_03, Constants.RECIPIENT),
    new QName(Constants.NS_URI_ADDRESSING_2004_03, Constants.RELATES_TO),
    new QName(Constants.NS_URI_ADDRESSING_2004_03, Constants.REPLY_TO),
    new QName(Constants.NS_URI_ADDRESSING_2004_03, Constants.TO),
    new QName(Constants.NS_URI_ADDRESSING_2004_08, Constants.ACTION),
    new QName(Constants.NS_URI_ADDRESSING_2004_08, Constants.FAULT_TO),
    new QName(Constants.NS_URI_ADDRESSING_2004_08, Constants.FROM),
    new QName(Constants.NS_URI_ADDRESSING_2004_08, Constants.MESSAGE_ID),
    new QName(Constants.NS_URI_ADDRESSING_2004_08, Constants.RECIPIENT),
    new QName(Constants.NS_URI_ADDRESSING_2004_08, Constants.RELATES_TO),
    new QName(Constants.NS_URI_ADDRESSING_2004_08, Constants.REPLY_TO),
    new QName(Constants.NS_URI_ADDRESSING_2004_08, Constants.TO)
  };

  /**
   * Constant to be returned by handleRequest(), handleResponse(), and handleFault().
   */
  protected static final boolean CONTINUE_HANDLER_CHAIN_PROCESSING = true;

  /**
   * Class logger.
   */
  private static final Log LOG = LogFactory.getLog(AbstractAddressingHandler.class);

  /**
   * Handler configuration.
   */
  private Map<String, String> handlerConfigProps;

  /**
   * Reference property names to process.
   */
  private List<QName> refPropQNames;

  /**
   * Returns a list QNames of the SOAP headers that this handler processes.
   *
   * @return Returns a list QNames of the SOAP headers that this handler processes
   */
  public QName[] getHeaders() {
    return UNDERSTOOD_HEADERS.clone();
  }

  /**
   * Initializes this handler. This handler supports the following configuration
   * properties:
   * <ul>
   *   <li>actor (optional) - only SOAP headers with this actor URI will be processed</li>
   *   <li>referencePropertyNames (optional) - a list of QNames of reference properties
   *       (i.e. - non-WSA SOAP headers) that should be processed; if set to "*", all
   *       properties will be processed, if not specified, only WSA headers will be
   *       processed</li>
   *   <li>removeHeaders (optional) - if set to "true", headers will be removed from the
   *       SOAP envelope after they have been processed</li>
   * </ul>
   *
   * @param handlerInfo
   *            handler info containing handler configuration properties from deployment
   *            descriptor
   */
  @SuppressWarnings("unchecked")
  @Override
  public void init(HandlerInfo handlerInfo) {
    super.init(handlerInfo);
    LOG.info("Initializing JAX-RPC handler " + this.getClass().getName() + "...");
    handlerConfigProps = handlerInfo.getHandlerConfig();
    initReferencePropertyQNames();
  }

  /**
   * Returns this handler's configuration properties.
   *
   * @return this handler's configuration properties
   */
  protected Map<String, String> getHandlerConfigProperties() {
    return handlerConfigProps;
  }

  /**
   * Returns the value of the specified handler configuration property.
   *
   * @param propName
   *            the name of a configuration property
   *
   * @return the value of the specified handler configuration property
   */
  protected String getHandlerConfigProperty(String propName) {
    return (String) getHandlerConfigProperties().get(propName);
  }

  /**
   * Returns true if the value of the specified handler configuration property is "true",
   * or false otherwise.
   *
   * @param propName
   *            the name of a configuration property
   *
   * @return true if the value of the specified handler configuration property is "true",
   *         or false otherwise
   */
  protected boolean isHandlerConfigPropertyTrue(String propName) {
    return Boolean.TRUE.toString().equalsIgnoreCase(getHandlerConfigProperty(propName));
  }

  /**
   * This method uses the Java UUID Generator library found through discovery.
   * <p>
   * Subclasses can optionally override it to use a different UUID generator.
   *
   * @return A unique identifier
   */
  protected String generateUUId() {
    try {
      return UUIdGeneratorFactory.getGenerator().generateUUId();
    } catch (RuntimeException re) {
      throw new JAXRPCException("Failed to generate UUId using default UUId generator",
          re);
    }
  }

  /**
   * Gets the value of the "MustUnderstandEnabled" field.
   *
   * @param msgContext Context to get the information from
   * @return Field value
   */
  protected boolean isMustUnderstandEnabled(MessageContext msgContext) {
    return isPropertyTrue(msgContext, Constants.ENV_ADDRESSING_SET_MUST_UNDERSTAND);
  }

  /**
   * Gets if the specified property has a "true" as its value.
   *
   * @param msgContext Context to get the information from
   * @param propName   Property name
   * @return true if the property is set to "true"
   */
  protected boolean isPropertyTrue(MessageContext msgContext, String propName) {
    return Boolean.TRUE.toString().equals(msgContext.getProperty(propName));
  }

  /**
   * Creates a new random message identifier.
   *
   * @return The message identifier URI
   * @throws URI.MalformedURIException If an invalid URI is created
   */
  protected MessageID createMessageID() throws URI.MalformedURIException {
    return new MessageID(new URI(UUID_URI_SCHEME + ":" + generateUUId()));
  }

  /**
   * Gets the list of processed property names.
   *
   * @return The list of property names
   */
  protected List<QName> getReferencePropertyQNames() {
    return refPropQNames;
  }

  /**
   * Returns the value of the "actor" handler config property.
   *
   * @return the value of the "actor" handler config property
   */
  protected String getActor() {
    return getHandlerConfigProperty(CONFIG_PROP_ACTOR);
  }

  /**
   * Returns the value of the "removeHeaders" handler config property.
   *
   * @return the value of the "removeHeaders" handler config property
   */
  protected boolean isRemoveHeadersEnabled() {
    return isHandlerConfigPropertyTrue(CONFIG_PROP_REMOVE_HEADERS);
  }

  /**
   * Retrieves QNames for reference properties from WSDD and initializes
   * refPropQNames - the list of reference properties that this handler
   * should care about.
   */
  protected void initReferencePropertyQNames() {
    String refPropNamesConfigPropValue
      = getHandlerConfigProperty(CONFIG_PROP_REFERENCE_PROPERTY_NAMES);

    if (refPropNamesConfigPropValue == null) {
      LOG.debug("No reference properties will be processed.");
      refPropQNames = new ArrayList<QName>();
    } else if (refPropNamesConfigPropValue.equals("*")) {
      LOG.debug("All reference properties will be processed.");
      refPropQNames = null;
    } else {
      refPropQNames = new ArrayList<QName>();
      StringTokenizer tokenizer = new StringTokenizer(refPropNamesConfigPropValue, ",");
      while (tokenizer.hasMoreTokens()) {
        String qnameString = tokenizer.nextToken().trim();
        try {
          QName qname = QName.valueOf(qnameString);
          LOG.debug("Reference properties with QName '" + qnameString
              + "' will be processed.");
          refPropQNames.add(qname);
        } catch (IllegalArgumentException iae) {
          LOG.warn("Invalid QName '" + qnameString + "' specified in handler config.");
        }
      }
    }
  }

}
