/*
 *  Copyright (c) 2008 Rodrigo Ruiz
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.apache.ws.addressing.handler;

import java.io.IOException;

import javax.xml.rpc.JAXRPCException;
import javax.xml.rpc.handler.MessageContext;
import javax.xml.rpc.handler.soap.SOAPMessageContext;
import javax.xml.soap.SOAPException;
import javax.xml.soap.SOAPMessage;

import org.apache.axis.message.addressing.Action;
import org.apache.axis.message.addressing.AddressingHeaders;
import org.apache.axis.message.addressing.Constants;
import org.apache.axis.message.addressing.EndpointReference;
import org.apache.axis.message.addressing.To;
import org.apache.axis.message.addressing.util.AddressingUtils;
import org.apache.axis.types.URI.MalformedURIException;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * A client-side JAX-RPC {@link javax.xml.rpc.handler.Handler} that inserts
 * WS-Addressing headers into outgoing SOAP requests
 * and extracts them from incoming SOAP responses.
 *
 * @author Davanum Srinivas
 * @author Ian P. Springer
 * @version $Revision: 14 $
 */
public class ClientSideAddressingHandler extends AbstractAddressingHandler {
  /**
   * Class logger.
   */
  private static final Log LOG = LogFactory.getLog(ClientSideAddressingHandler.class);

  /**
   * SOAP Action property name.
   */
  private static final String MSG_CONTEXT_PROP_SOAP_ACTION
    = javax.xml.rpc.Call.SOAPACTION_URI_PROPERTY;

  /**
   * {@inheritDoc}
   */
  public boolean handleRequest(MessageContext msgContext) {
    SOAPMessageContext soapMsgContext = (SOAPMessageContext) msgContext;
    try {
      AddressingHeaders headers = (AddressingHeaders) msgContext
      .getProperty(Constants.ENV_ADDRESSING_REQUEST_HEADERS);
      if (headers == null) {
        headers = new AddressingHeaders();
      }
      headers.setSetMustUnderstand(isMustUnderstandEnabled(msgContext));
      addMessageIdHeader(headers);
      addToHeader(headers, msgContext);
      addActionHeader(headers, msgContext);
      addFromHeader(headers, msgContext);
      addReplyToHeader(headers, msgContext);
      if (headers.getFaultTo() == null
          && msgContext.containsProperty(Constants.ENV_ADDRESSING_FAULTTO_URI)) {
        headers.setFaultTo(new EndpointReference((String) msgContext
            .getProperty(Constants.ENV_ADDRESSING_FAULTTO_URI)));
      }

      SOAPMessage msg = soapMsgContext.getMessage();
      headers.toEnvelope(msg.getSOAPPart().getEnvelope(), getActor());
      msgContext.setProperty(Constants.ENV_ADDRESSING_REQUEST_HEADERS, headers);
    } catch (IOException e) {
      if (LOG.isDebugEnabled()) {
        LOG.debug("Unexpected error in handleRequest()", e);
      }
      throw new JAXRPCException("unexpected error in handleRequest()", e);
    } catch (SOAPException e) {
      if (LOG.isDebugEnabled()) {
        LOG.debug("Unexpected error in handleRequest()", e);
      }
      throw new JAXRPCException("unexpected error in handleRequest()", e);
    }
    return CONTINUE_HANDLER_CHAIN_PROCESSING;
  }

  /**
   * {@inheritDoc}
   */
  public boolean handleResponse(MessageContext msgContext) {
    SOAPMessageContext soapMsgContext = (SOAPMessageContext) msgContext;
    try {
      SOAPMessage msg = soapMsgContext.getMessage();
      if (msg == null) {
        return CONTINUE_HANDLER_CHAIN_PROCESSING;
      }

      AddressingHeaders headers = new AddressingHeaders(
        msg.getSOAPPart().getEnvelope(), getActor(), true, isRemoveHeadersEnabled(),
        false, getReferencePropertyQNames());
      msgContext.setProperty(Constants.ENV_ADDRESSING_RESPONSE_HEADERS, headers);
    } catch (Exception e) {
      if (LOG.isDebugEnabled()) {
        e.printStackTrace();
      }
      throw new JAXRPCException("unexpected error in handleResponse(): " + e, e);
    }
    return CONTINUE_HANDLER_CHAIN_PROCESSING;
  }

  /**
   * {@inheritDoc}
   */
  public boolean handleFault(MessageContext messageContext) {
    return CONTINUE_HANDLER_CHAIN_PROCESSING;
  }

  /**
   * The JAX-RPC APIs don't provide a way to retrieve the value of the SOAPAction
   * HTTP header, so platform-specific subclasses should implement this method if
   * feasible.
   *
   * @param msgContext JAX-RPC message context
   * @return the value of the SOAPAction HTTP header
   */
  protected String getSOAPAction(MessageContext msgContext) {
    return (String) msgContext.getProperty(MSG_CONTEXT_PROP_SOAP_ACTION);
  }

  /**
   * The JAX-RPC APIs don't provide a way to set the value of the SOAPAction HTTP header,
   * so platform-specific subclasses should implement this method if feasible.
   *
   * @param msgContext JAX-RPC message context
   * @param actionURI  the SOAPAction URI to be set
   */
  protected void setSOAPAction(MessageContext msgContext, String actionURI) {
    msgContext.setProperty(MSG_CONTEXT_PROP_SOAP_ACTION, actionURI);
  }

  /**
   * Returns the endpoint URL for the specified message context. Since there's no
   * generic JAX-RPC way to obtain the request's endpoint URL, we return null here.
   * Platform-specific handlers can override if the platform provides a means to
   * obtain the endpoint URL.
   *
   * @param msgContext Context information
   * @return the endpoint URL for the specified message context
   */
  protected String getEndpointURL(MessageContext msgContext) {
    return null;
  }

  /**
   * Builds a ReplyTo header and adds it to the headers instance.
   *
   * @param addrHeaders Headers container
   * @param msgContext  Context information
   * @throws MalformedURIException If an invalid URI is built
   */
  private void addReplyToHeader(AddressingHeaders addrHeaders, MessageContext msgContext)
    throws MalformedURIException {
    if (isPropertyTrue(msgContext, Constants.ENV_ADDRESSING_SEND_REPLYTO)
        && addrHeaders.getReplyTo() == null) {
      if (msgContext.containsProperty(Constants.ENV_ADDRESSING_REPLYTO_URI)) {
        addrHeaders.setReplyTo(new EndpointReference((String) msgContext
            .getProperty(Constants.ENV_ADDRESSING_REPLYTO_URI)));
      } else {
        addrHeaders.setReplyTo(addrHeaders.getFrom());
      }
    }
  }

  /**
   * Builds a From header and adds it to the specified headers container.
   *
   * @param addrHeaders Container
   * @param ctx  Context information
   * @throws MalformedURIException If an invalid URI is built
   */
  private void addFromHeader(AddressingHeaders addrHeaders, MessageContext ctx)
    throws MalformedURIException {
    if (addrHeaders.getFrom() == null) {
      String uri = (String)ctx.getProperty(Constants.ENV_ADDRESSING_FROM_URI);
      if (uri == null) {
        uri = AddressingUtils.getAnonymousRoleURI(ctx);
      }
      addrHeaders.setFrom(new EndpointReference(uri));
    }
  }

  /**
   * Builds a MessageID header and adds it to the specified headers container.
   *
   * @param addrHeaders Container
   * @throws MalformedURIException If an invalid URI is built
   */
  private void addMessageIdHeader(AddressingHeaders addrHeaders)
    throws MalformedURIException {
    if (addrHeaders.getMessageID() == null) {
      addrHeaders.setMessageID(createMessageID());
    }
  }

  /**
   * Builds a To header and adds it to the specified headers container.
   *
   * @param addrHeaders Container
   * @param msgContext  Context information
   * @throws MalformedURIException If an invalid URI is built
   */
  private void addToHeader(AddressingHeaders addrHeaders, MessageContext msgContext)
    throws MalformedURIException {
    if (addrHeaders.getTo() == null) {
      String endpointURL = getEndpointURL(msgContext);
      addrHeaders.setTo(endpointURL != null ? new To(endpointURL) : null);
    }
  }

  /**
   * Builds an Action header and adds it to the specified headers container.
   *
   * @param addrHeaders Container
   * @param msgContext  Context information
   * @throws MalformedURIException If an invalid URI is built
   */
  private void addActionHeader(AddressingHeaders addrHeaders, MessageContext msgContext)
    throws MalformedURIException {
    String actionURI = getSOAPAction(msgContext);
    if (actionURI != null) {
      addrHeaders.setAction(new Action(actionURI));
    } else if (addrHeaders.getAction() != null) {
      setSOAPAction(msgContext, addrHeaders.getAction().toString());
    }
  }
}
