/*
 *  Copyright (c) 2008 Rodrigo Ruiz
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package org.apache.ws.addressing.uuid;

import javax.xml.rpc.JAXRPCException;
import java.lang.reflect.Method;

/**
 * A {@link UUIdGenerator} implementation that uses JUG, an open source Java UUID
 * generation library. Reflection is utilized to avoid a compile-time dependency
 * on jug.jar.
 *
 * @author Ian P. Springer
 * @author Rodrigo Ruiz
 */
public class JugUUIdGenerator implements UUIdGenerator {
  /**
   * JUG Generator class name.
   */
  private static final String JUG_UUIDGENERATOR_CLASSNAME
    = "org.doomdark.uuid.UUIDGenerator";

  /**
   * Cached Method.
   */
  private final Method generateUUIdMethod;

  /**
   * Cached Method.
   */
  private final Class<?> jugUuidGenClass;

  /**
   * Actual generator.
   */
  private final Object gen;

  /**
   * Creates an instance.
   */
  public JugUUIdGenerator() {
    try {
      jugUuidGenClass = Class.forName(JUG_UUIDGENERATOR_CLASSNAME);
    } catch (ClassNotFoundException e) {
      throw new RuntimeException("Java UUID Generator (JUG) not found.", e);
    }
    try {
      generateUUIdMethod = jugUuidGenClass.getMethod("generateTimeBasedUUID");
    } catch (Exception e) {
      throw new RuntimeException("Error initializing Java UUID Generator (JUG)", e);
    }

    try {
      gen =  jugUuidGenClass.newInstance();
    } catch (Throwable t) {
      throw new RuntimeException("Error instantiating Java UUID Generator (JUG)", t);
    }
  }

  /**
   * {@inheritDoc}
   */
  public synchronized String generateUUId() {
    try {
      return generateUUIdMethod.invoke(gen).toString();
    } catch (Exception e) {
      throw new JAXRPCException(e);
    }
  }
}
